"""Properties panel for selected monitor using Adw.PreferencesPage."""

from __future__ import annotations

import gi
gi.require_version("Gtk", "4.0")
gi.require_version("Adw", "1")
from gi.repository import Gtk, Adw, GLib, GObject

from .models import (
    MonitorConfig, ResolutionMode, PositionMode, ScaleMode,
    Transform, VRR,
)


def _fix_spin_icons(widget: Gtk.Widget) -> None:
    """Replace SpinRow +/- button icons with text labels after a delay."""
    _LABELS = {
        "list-add-symbolic": "+",
        "list-remove-symbolic": "\u2212",
        "value-increase-symbolic": "+",
        "value-decrease-symbolic": "\u2212",
    }

    def _do_fix():
        _walk(widget)
        return False  # don't repeat

    def _walk(w):
        child = w.get_first_child()
        while child:
            next_s = child.get_next_sibling()
            if isinstance(child, Gtk.Button):
                icon = child.get_icon_name() or ""
                if not icon:
                    bc = child.get_child()
                    if isinstance(bc, Gtk.Image):
                        icon = bc.get_icon_name() or ""
                if icon in _LABELS:
                    child.set_icon_name("")
                    lbl = Gtk.Label(label=_LABELS[icon])
                    lbl.add_css_class("heading")
                    child.set_child(lbl)
            # Always recurse into all children
            _walk(child)
            child = next_s

    GLib.timeout_add(150, _do_fix)


class PropertiesPanel(Adw.PreferencesPage):
    """Side panel showing properties of the selected monitor."""

    __gtype_name__ = "PropertiesPanel"

    __gsignals__ = {
        "property-changed": (GObject.SignalFlags.RUN_FIRST, None, ()),
    }

    def __init__(self) -> None:
        super().__init__()
        self._monitor: MonitorConfig | None = None
        self._building = False

        self._build_ui()

    def _build_ui(self) -> None:
        # ── Monitor Info ─────────────────────────────────────────────
        grp_info = Adw.PreferencesGroup(title="Monitor")
        self.add(grp_info)

        self._row_name = Adw.ActionRow(title="Name", icon_name="video-display-symbolic")
        self._lbl_name = Gtk.Label(label="—", xalign=1)
        self._lbl_name.add_css_class("dim-label")
        self._row_name.add_suffix(self._lbl_name)
        grp_info.add(self._row_name)

        self._row_desc = Adw.ActionRow(title="Description", icon_name="text-x-generic-symbolic")
        self._lbl_desc = Gtk.Label(label="—", xalign=1, wrap=True, max_width_chars=24)
        self._lbl_desc.add_css_class("dim-label")
        self._row_desc.add_suffix(self._lbl_desc)
        grp_info.add(self._row_desc)

        self._sw_enabled = Adw.SwitchRow(title="Enabled", icon_name="system-shutdown-symbolic")
        self._sw_enabled.connect("notify::active", self._on_changed)
        grp_info.add(self._sw_enabled)

        # ── Resolution ───────────────────────────────────────────────
        grp_res = Adw.PreferencesGroup(title="Resolution")
        self.add(grp_res)

        self._combo_res_mode = Adw.ComboRow(title="Mode", icon_name="preferences-other-symbolic")
        modes = Gtk.StringList.new([m.value for m in ResolutionMode])
        self._combo_res_mode.set_model(modes)
        self._combo_res_mode.connect("notify::selected", self._on_res_mode_changed)
        grp_res.add(self._combo_res_mode)

        self._combo_resolution = Adw.ComboRow(title="Resolution", icon_name="preferences-desktop-display-symbolic")
        self._combo_resolution.connect("notify::selected", self._on_resolution_changed)
        grp_res.add(self._combo_resolution)

        self._spin_width = Adw.SpinRow.new_with_range(320, 15360, 1)
        self._spin_width.set_title("Width")
        self._spin_width.connect("notify::value", self._on_changed)
        grp_res.add(self._spin_width)
        _fix_spin_icons(self._spin_width)

        self._spin_height = Adw.SpinRow.new_with_range(200, 8640, 1)
        self._spin_height.set_title("Height")
        self._spin_height.connect("notify::value", self._on_changed)
        grp_res.add(self._spin_height)
        _fix_spin_icons(self._spin_height)

        self._spin_refresh = Adw.SpinRow.new_with_range(1, 600, 0.01)
        self._spin_refresh.set_title("Refresh Rate")
        self._spin_refresh.set_digits(2)
        self._spin_refresh.connect("notify::value", self._on_changed)
        grp_res.add(self._spin_refresh)
        _fix_spin_icons(self._spin_refresh)

        # ── Position ─────────────────────────────────────────────────
        grp_pos = Adw.PreferencesGroup(title="Position")
        self.add(grp_pos)

        self._combo_pos_mode = Adw.ComboRow(title="Mode", icon_name="preferences-other-symbolic")
        pos_modes = Gtk.StringList.new([m.value for m in PositionMode])
        self._combo_pos_mode.set_model(pos_modes)
        self._combo_pos_mode.connect("notify::selected", self._on_pos_mode_changed)
        grp_pos.add(self._combo_pos_mode)

        self._spin_x = Adw.SpinRow.new_with_range(-32768, 32768, 1)
        self._spin_x.set_title("X")
        self._spin_x.connect("notify::value", self._on_changed)
        grp_pos.add(self._spin_x)
        _fix_spin_icons(self._spin_x)

        self._spin_y = Adw.SpinRow.new_with_range(-32768, 32768, 1)
        self._spin_y.set_title("Y")
        self._spin_y.connect("notify::value", self._on_changed)
        grp_pos.add(self._spin_y)
        _fix_spin_icons(self._spin_y)

        # ── Scale & Transform ────────────────────────────────────────
        grp_scale = Adw.PreferencesGroup(title="Scale &amp; Transform")
        self.add(grp_scale)

        self._combo_scale_mode = Adw.ComboRow(title="Scale Mode")
        scale_modes = Gtk.StringList.new([m.value for m in ScaleMode])
        self._combo_scale_mode.set_model(scale_modes)
        self._combo_scale_mode.connect("notify::selected", self._on_scale_mode_changed)
        grp_scale.add(self._combo_scale_mode)

        self._spin_scale = Adw.SpinRow.new_with_range(0.1, 10.0, 0.05)
        self._spin_scale.set_title("Scale")
        self._spin_scale.set_digits(2)
        self._spin_scale.connect("notify::value", self._on_changed)
        grp_scale.add(self._spin_scale)
        _fix_spin_icons(self._spin_scale)

        self._combo_transform = Adw.ComboRow(title="Transform", icon_name="object-rotate-right-symbolic")
        transforms = Gtk.StringList.new([t.label for t in Transform])
        self._combo_transform.set_model(transforms)
        self._combo_transform.connect("notify::selected", self._on_changed)
        grp_scale.add(self._combo_transform)

        # ── Advanced ─────────────────────────────────────────────────
        grp_adv = Adw.PreferencesGroup(title="Advanced")
        self.add(grp_adv)

        self._combo_mirror = Adw.ComboRow(title="Mirror Of", icon_name="edit-copy-symbolic")
        self._combo_mirror.connect("notify::selected", self._on_changed)
        grp_adv.add(self._combo_mirror)

        self._combo_bitdepth = Adw.ComboRow(title="Bit Depth", icon_name="color-select-symbolic")
        self._combo_bitdepth.set_model(Gtk.StringList.new(["8", "10"]))
        self._combo_bitdepth.connect("notify::selected", self._on_changed)
        grp_adv.add(self._combo_bitdepth)

        self._combo_vrr = Adw.ComboRow(title="VRR", icon_name="display-brightness-symbolic")
        self._combo_vrr.set_model(Gtk.StringList.new(["Off", "On", "Fullscreen only"]))
        self._combo_vrr.connect("notify::selected", self._on_changed)
        grp_adv.add(self._combo_vrr)

        self._combo_cm = Adw.ComboRow(title="Color Management", icon_name="applications-graphics-symbolic")
        cm_options = ["None", "auto", "srgb", "dcip3", "dp3", "adobe", "wide", "edid", "hdr", "hdredid"]
        self._combo_cm.set_model(Gtk.StringList.new(cm_options))
        self._combo_cm.connect("notify::selected", self._on_changed)
        grp_adv.add(self._combo_cm)

        self._spin_sdr_bright = Adw.SpinRow.new_with_range(0.0, 5.0, 0.05)
        self._spin_sdr_bright.set_title("SDR Brightness")
        self._spin_sdr_bright.set_digits(2)
        self._spin_sdr_bright.connect("notify::value", self._on_changed)
        grp_adv.add(self._spin_sdr_bright)
        _fix_spin_icons(self._spin_sdr_bright)

        self._spin_sdr_sat = Adw.SpinRow.new_with_range(0.0, 5.0, 0.05)
        self._spin_sdr_sat.set_title("SDR Saturation")
        self._spin_sdr_sat.set_digits(2)
        self._spin_sdr_sat.connect("notify::value", self._on_changed)
        grp_adv.add(self._spin_sdr_sat)
        _fix_spin_icons(self._spin_sdr_sat)

        # ── Reserved Area ────────────────────────────────────────────
        grp_reserved = Adw.PreferencesGroup(title="Reserved Area")
        self.add(grp_reserved)

        self._spin_res_top = Adw.SpinRow.new_with_range(0, 500, 1)
        self._spin_res_top.set_title("Top")
        self._spin_res_top.connect("notify::value", self._on_changed)
        grp_reserved.add(self._spin_res_top)
        _fix_spin_icons(self._spin_res_top)

        self._spin_res_bottom = Adw.SpinRow.new_with_range(0, 500, 1)
        self._spin_res_bottom.set_title("Bottom")
        self._spin_res_bottom.connect("notify::value", self._on_changed)
        grp_reserved.add(self._spin_res_bottom)
        _fix_spin_icons(self._spin_res_bottom)

        self._spin_res_left = Adw.SpinRow.new_with_range(0, 500, 1)
        self._spin_res_left.set_title("Left")
        self._spin_res_left.connect("notify::value", self._on_changed)
        grp_reserved.add(self._spin_res_left)
        _fix_spin_icons(self._spin_res_left)

        self._spin_res_right = Adw.SpinRow.new_with_range(0, 500, 1)
        self._spin_res_right.set_title("Right")
        self._spin_res_right.connect("notify::value", self._on_changed)
        grp_reserved.add(self._spin_res_right)
        _fix_spin_icons(self._spin_res_right)

    def set_mirror_monitors(self, names: list[str]) -> None:
        """Update the mirror dropdown with available monitor names."""
        self._building = True
        options = ["None"] + [n for n in names if n != (self._monitor.name if self._monitor else "")]
        self._combo_mirror.set_model(Gtk.StringList.new(options))
        self._building = False

    def update_from_monitor(self, monitor: MonitorConfig | None) -> None:
        """Populate all fields from a MonitorConfig."""
        self._building = True
        self._monitor = monitor

        if monitor is None:
            self.set_sensitive(False)
            self._building = False
            return

        self.set_sensitive(True)

        self._lbl_name.set_label(monitor.name or "—")
        self._lbl_desc.set_label(monitor.description or "—")
        self._sw_enabled.set_active(monitor.enabled)

        # Resolution mode
        res_modes = [m.value for m in ResolutionMode]
        idx = res_modes.index(monitor.resolution_mode.value) if monitor.resolution_mode.value in res_modes else 0
        self._combo_res_mode.set_selected(idx)

        # Available modes dropdown
        if monitor.available_modes:
            self._combo_resolution.set_model(Gtk.StringList.new(monitor.available_modes))
            # Find closest match
            for i, mode in enumerate(monitor.available_modes):
                if mode.startswith(f"{monitor.width}x{monitor.height}"):
                    self._combo_resolution.set_selected(i)
                    break
        else:
            self._combo_resolution.set_model(Gtk.StringList.new([]))

        self._spin_width.set_value(monitor.width)
        self._spin_height.set_value(monitor.height)
        self._spin_refresh.set_value(monitor.refresh_rate)

        # Show/hide manual fields based on mode
        is_explicit = monitor.resolution_mode == ResolutionMode.EXPLICIT
        self._combo_resolution.set_visible(is_explicit)
        self._spin_width.set_visible(is_explicit)
        self._spin_height.set_visible(is_explicit)
        self._spin_refresh.set_visible(is_explicit)

        # Position mode
        pos_modes = [m.value for m in PositionMode]
        idx = pos_modes.index(monitor.position_mode.value) if monitor.position_mode.value in pos_modes else 0
        self._combo_pos_mode.set_selected(idx)

        self._spin_x.set_value(monitor.x)
        self._spin_y.set_value(monitor.y)
        is_explicit_pos = monitor.position_mode == PositionMode.EXPLICIT
        self._spin_x.set_visible(is_explicit_pos)
        self._spin_y.set_visible(is_explicit_pos)

        # Scale
        scale_modes = [m.value for m in ScaleMode]
        idx = scale_modes.index(monitor.scale_mode.value) if monitor.scale_mode.value in scale_modes else 0
        self._combo_scale_mode.set_selected(idx)
        self._spin_scale.set_value(monitor.scale)
        self._spin_scale.set_visible(monitor.scale_mode == ScaleMode.EXPLICIT)

        # Transform
        self._combo_transform.set_selected(monitor.transform.value)

        # Mirror
        model = self._combo_mirror.get_model()
        if model:
            if monitor.mirror_of:
                for i in range(model.get_n_items()):
                    if model.get_string(i) == monitor.mirror_of:
                        self._combo_mirror.set_selected(i)
                        break
            else:
                self._combo_mirror.set_selected(0)

        # Advanced
        self._combo_bitdepth.set_selected(0 if monitor.bitdepth == 8 else 1)
        self._combo_vrr.set_selected(monitor.vrr.value)

        cm_options = ["None", "auto", "srgb", "dcip3", "dp3", "adobe", "wide", "edid", "hdr", "hdredid"]
        cm_val = monitor.color_management or "None"
        idx = cm_options.index(cm_val) if cm_val in cm_options else 0
        self._combo_cm.set_selected(idx)

        self._spin_sdr_bright.set_value(monitor.sdr_brightness)
        self._spin_sdr_sat.set_value(monitor.sdr_saturation)

        # Reserved
        self._spin_res_top.set_value(monitor.reserved_top)
        self._spin_res_bottom.set_value(monitor.reserved_bottom)
        self._spin_res_left.set_value(monitor.reserved_left)
        self._spin_res_right.set_value(monitor.reserved_right)

        self._building = False

    def _apply_to_monitor(self) -> None:
        """Write current UI values back to the MonitorConfig."""
        m = self._monitor
        if m is None:
            return

        m.enabled = self._sw_enabled.get_active()

        # Resolution mode
        res_modes = list(ResolutionMode)
        m.resolution_mode = res_modes[self._combo_res_mode.get_selected()]

        if m.resolution_mode == ResolutionMode.EXPLICIT:
            # Check if a mode was selected from dropdown
            sel = self._combo_resolution.get_selected()
            model = self._combo_resolution.get_model()
            if model and sel < model.get_n_items() and sel != Gtk.INVALID_LIST_POSITION:
                mode_str = model.get_string(sel)
                # Parse "WxH@R" or "WxH@R Hz"
                self._parse_mode_string(m, mode_str)
            else:
                m.width = int(self._spin_width.get_value())
                m.height = int(self._spin_height.get_value())
                m.refresh_rate = round(self._spin_refresh.get_value(), 2)

        # Position
        pos_modes = list(PositionMode)
        m.position_mode = pos_modes[self._combo_pos_mode.get_selected()]
        if m.position_mode == PositionMode.EXPLICIT:
            m.x = int(self._spin_x.get_value())
            m.y = int(self._spin_y.get_value())

        # Scale
        scale_modes = list(ScaleMode)
        m.scale_mode = scale_modes[self._combo_scale_mode.get_selected()]
        if m.scale_mode == ScaleMode.EXPLICIT:
            m.scale = round(self._spin_scale.get_value(), 2)

        # Transform
        m.transform = Transform(self._combo_transform.get_selected())

        # Mirror
        mirror_model = self._combo_mirror.get_model()
        if mirror_model:
            sel = self._combo_mirror.get_selected()
            if sel == 0 or sel == Gtk.INVALID_LIST_POSITION:
                m.mirror_of = ""
            else:
                m.mirror_of = mirror_model.get_string(sel)

        # Advanced
        m.bitdepth = 10 if self._combo_bitdepth.get_selected() == 1 else 8
        m.vrr = VRR(self._combo_vrr.get_selected())

        cm_options = ["", "auto", "srgb", "dcip3", "dp3", "adobe", "wide", "edid", "hdr", "hdredid"]
        m.color_management = cm_options[self._combo_cm.get_selected()]

        m.sdr_brightness = round(self._spin_sdr_bright.get_value(), 2)
        m.sdr_saturation = round(self._spin_sdr_sat.get_value(), 2)

        # Reserved
        m.reserved_top = int(self._spin_res_top.get_value())
        m.reserved_bottom = int(self._spin_res_bottom.get_value())
        m.reserved_left = int(self._spin_res_left.get_value())
        m.reserved_right = int(self._spin_res_right.get_value())

    def _parse_mode_string(self, m: MonitorConfig, mode_str: str) -> None:
        """Parse a mode string like '1920x1080@60.00' into monitor fields."""
        try:
            res_part, rate_part = mode_str.split("@")
            w, h = res_part.split("x")
            m.width = int(w)
            m.height = int(h)
            # Remove trailing " Hz" if present
            rate_part = rate_part.replace("Hz", "").strip()
            m.refresh_rate = round(float(rate_part), 2)
        except (ValueError, IndexError):
            pass

    def _on_changed(self, *args) -> None:
        if self._building or self._monitor is None:
            return
        self._apply_to_monitor()
        self.emit("property-changed")

    def _on_res_mode_changed(self, *args) -> None:
        if self._building:
            return
        res_modes = list(ResolutionMode)
        mode = res_modes[self._combo_res_mode.get_selected()]
        is_explicit = mode == ResolutionMode.EXPLICIT
        self._combo_resolution.set_visible(is_explicit)
        self._spin_width.set_visible(is_explicit)
        self._spin_height.set_visible(is_explicit)
        self._spin_refresh.set_visible(is_explicit)
        self._on_changed()

    def _on_resolution_changed(self, *args) -> None:
        if self._building or self._monitor is None:
            return
        sel = self._combo_resolution.get_selected()
        model = self._combo_resolution.get_model()
        if model and sel < model.get_n_items() and sel != Gtk.INVALID_LIST_POSITION:
            mode_str = model.get_string(sel)
            self._building = True
            self._parse_mode_string(self._monitor, mode_str)
            self._spin_width.set_value(self._monitor.width)
            self._spin_height.set_value(self._monitor.height)
            self._spin_refresh.set_value(self._monitor.refresh_rate)
            self._building = False
            self._on_changed()

    def _on_pos_mode_changed(self, *args) -> None:
        if self._building:
            return
        pos_modes = list(PositionMode)
        mode = pos_modes[self._combo_pos_mode.get_selected()]
        is_explicit = mode == PositionMode.EXPLICIT
        self._spin_x.set_visible(is_explicit)
        self._spin_y.set_visible(is_explicit)
        self._on_changed()

    def _on_scale_mode_changed(self, *args) -> None:
        if self._building:
            return
        scale_modes = list(ScaleMode)
        mode = scale_modes[self._combo_scale_mode.get_selected()]
        self._spin_scale.set_visible(mode == ScaleMode.EXPLICIT)
        self._on_changed()



"""Application entry point."""

from __future__ import annotations

import sys

import gi
gi.require_version("Gtk", "4.0")
gi.require_version("Adw", "1")
from gi.repository import Adw, Gio

from .utils import APP_ID


class MonitorApp(Adw.Application):
    """Main application class."""

    def __init__(self) -> None:
        super().__init__(
            application_id=APP_ID,
            flags=Gio.ApplicationFlags.DEFAULT_FLAGS,
        )

    def do_activate(self) -> None:
        win = self.get_active_window()
        if win is None:
            from .window import MainWindow
            win = MainWindow(self)
        win.present()


def main() -> None:
    app = MonitorApp()
    app.run(sys.argv)



"""Profile management: save, load, delete, list, and match profiles."""

from __future__ import annotations

from pathlib import Path

from .models import Profile
from .utils import profiles_dir, read_json, write_json


class ProfileManager:
    """Manages monitor configuration profiles stored as JSON files."""

    def __init__(self, directory: Path | None = None) -> None:
        self._dir = directory or profiles_dir()

    def _path_for(self, name: str) -> Path:
        safe = name.replace("/", "_").replace("\\", "_")
        return self._dir / f"{safe}.json"

    def save(self, profile: Profile) -> Path:
        """Save a profile to disk. Returns the file path."""
        path = self._path_for(profile.name)
        write_json(path, profile.to_dict())
        return path

    def load(self, name: str) -> Profile | None:
        """Load a profile by name."""
        data = read_json(self._path_for(name))
        if data is None:
            return None
        return Profile.from_dict(data)

    def delete(self, name: str) -> bool:
        """Delete a profile. Returns True if it existed."""
        path = self._path_for(name)
        if path.exists():
            path.unlink()
            return True
        return False

    def list_profiles(self) -> list[str]:
        """Return sorted list of profile names."""
        names = []
        for p in sorted(self._dir.glob("*.json")):
            data = read_json(p)
            if data and isinstance(data, dict):
                names.append(data.get("name", p.stem))
        return names

    def list_all(self) -> list[Profile]:
        """Load all profiles."""
        profiles = []
        for p in sorted(self._dir.glob("*.json")):
            data = read_json(p)
            if data and isinstance(data, dict):
                profiles.append(Profile.from_dict(data))
        return profiles

    def find_best_match(self, current_fingerprint: list[str]) -> Profile | None:
        """Find the best matching profile for the given monitor fingerprint.

        Uses exact match first, then Jaccard similarity >= 0.5 as fallback.
        """
        if not current_fingerprint:
            return None

        current_set = set(current_fingerprint)
        best_profile: Profile | None = None
        best_score = 0.0

        for profile in self.list_all():
            fp = profile.fingerprint
            if not fp:
                continue

            # Exact match
            if fp == current_fingerprint:
                return profile

            # Jaccard similarity
            profile_set = set(fp)
            intersection = len(current_set & profile_set)
            union = len(current_set | profile_set)
            if union == 0:
                continue
            score = intersection / union
            if score > best_score:
                best_score = score
                best_profile = profile

        if best_score >= 0.5:
            return best_profile
        return None



"""Monitor arrangement canvas using Gtk.DrawingArea + Cairo."""

from __future__ import annotations

import math

import gi
gi.require_version("Gtk", "4.0")
from gi.repository import Gtk, GObject

from .models import MonitorConfig, PositionMode


# Snap distance in canvas pixels
SNAP_DISTANCE = 12
# Minimum scale factor for zoom
MIN_ZOOM = 0.05
MAX_ZOOM = 3.0

# Colors
COLOR_BG = (0.12, 0.12, 0.14)
COLOR_GRID = (0.18, 0.18, 0.20)
COLOR_MONITOR = (0.22, 0.24, 0.28)
COLOR_MONITOR_BORDER = (0.4, 0.42, 0.46)
COLOR_SELECTED = (0.26, 0.52, 0.96)
COLOR_SELECTED_FILL = (0.20, 0.35, 0.55)
COLOR_DISABLED = (0.5, 0.2, 0.2)
COLOR_TEXT = (0.9, 0.9, 0.92)
COLOR_TEXT_DIM = (0.6, 0.62, 0.64)


class MonitorCanvas(Gtk.DrawingArea):
    """Canvas widget for arranging monitors with drag-and-drop."""

    __gtype_name__ = "MonitorCanvas"

    __gsignals__ = {
        "monitor-selected": (GObject.SignalFlags.RUN_FIRST, None, (int,)),
        "monitor-moved": (GObject.SignalFlags.RUN_FIRST, None, (int,)),
        "monitor-double-clicked": (GObject.SignalFlags.RUN_FIRST, None, (int,)),
    }

    def __init__(self) -> None:
        super().__init__()
        self._monitors: list[MonitorConfig] = []
        self._use_description: bool = True
        self._selected: int = -1
        self._zoom: float = 0.15
        self._pan_x: float = 0.0
        self._pan_y: float = 0.0

        # Drag state
        self._dragging: bool = False
        self._drag_start_x: float = 0
        self._drag_start_y: float = 0
        self._drag_orig_mx: int = 0
        self._drag_orig_my: int = 0

        # Pan state
        self._panning: bool = False
        self._pan_start_x: float = 0
        self._pan_start_y: float = 0
        self._pan_orig_px: float = 0
        self._pan_orig_py: float = 0

        self.set_draw_func(self._draw)
        self.set_hexpand(True)
        self.set_vexpand(True)

        # Click gesture for selection
        click = Gtk.GestureClick()
        click.set_button(1)
        click.connect("pressed", self._on_click_pressed)
        self.add_controller(click)

        # Middle-click for panning
        mid_click = Gtk.GestureClick()
        mid_click.set_button(2)
        mid_click.connect("pressed", self._on_mid_pressed)
        self.add_controller(mid_click)

        # Drag gesture for moving monitors
        drag = Gtk.GestureDrag()
        drag.set_button(1)
        drag.connect("drag-begin", self._on_drag_begin)
        drag.connect("drag-update", self._on_drag_update)
        drag.connect("drag-end", self._on_drag_end)
        self.add_controller(drag)

        # Middle-button drag for panning
        pan_drag = Gtk.GestureDrag()
        pan_drag.set_button(2)
        pan_drag.connect("drag-begin", self._on_pan_begin)
        pan_drag.connect("drag-update", self._on_pan_update)
        self.add_controller(pan_drag)

        # Scroll for zoom
        scroll = Gtk.EventControllerScroll()
        scroll.set_flags(Gtk.EventControllerScrollFlags.VERTICAL)
        scroll.connect("scroll", self._on_scroll)
        self.add_controller(scroll)

    @property
    def monitors(self) -> list[MonitorConfig]:
        return self._monitors

    @monitors.setter
    def monitors(self, value: list[MonitorConfig]) -> None:
        self._monitors = value
        if self._selected >= len(value):
            self._selected = len(value) - 1 if value else -1
        self._auto_fit()
        self.queue_draw()

    @property
    def selected_index(self) -> int:
        return self._selected

    @selected_index.setter
    def selected_index(self, value: int) -> None:
        if value != self._selected:
            self._selected = value
            self.queue_draw()
            self.emit("monitor-selected", value)

    def set_use_description(self, val: bool) -> None:
        """Toggle between showing monitor descriptions vs port names."""
        self._use_description = val
        self.queue_draw()

    def _auto_fit(self) -> None:
        """Auto-fit zoom and pan to show all monitors."""
        if not self._monitors:
            return
        min_x = min(m.x for m in self._monitors)
        min_y = min(m.y for m in self._monitors)
        max_x = max(m.x + m.logical_width for m in self._monitors)
        max_y = max(m.y + m.logical_height for m in self._monitors)

        w = self.get_width() or 800
        h = self.get_height() or 600

        span_x = max_x - min_x
        span_y = max_y - min_y
        if span_x <= 0 or span_y <= 0:
            return

        margin = 80
        zoom_x = (w - margin * 2) / span_x
        zoom_y = (h - margin * 2) / span_y
        self._zoom = max(MIN_ZOOM, min(MAX_ZOOM, min(zoom_x, zoom_y)))

        cx = (min_x + max_x) / 2
        cy = (min_y + max_y) / 2
        self._pan_x = w / 2 - cx * self._zoom
        self._pan_y = h / 2 - cy * self._zoom

    def _screen_to_logical(self, sx: float, sy: float) -> tuple[float, float]:
        """Convert screen coordinates to logical monitor coordinates."""
        lx = (sx - self._pan_x) / self._zoom
        ly = (sy - self._pan_y) / self._zoom
        return lx, ly

    def _logical_to_screen(self, lx: float, ly: float) -> tuple[float, float]:
        """Convert logical coordinates to screen coordinates."""
        sx = lx * self._zoom + self._pan_x
        sy = ly * self._zoom + self._pan_y
        return sx, sy

    def _hit_test(self, sx: float, sy: float) -> int:
        """Return index of monitor at screen position, or -1."""
        lx, ly = self._screen_to_logical(sx, sy)
        # Check in reverse order (topmost first)
        for i in range(len(self._monitors) - 1, -1, -1):
            m = self._monitors[i]
            if (m.x <= lx <= m.x + m.logical_width and
                    m.y <= ly <= m.y + m.logical_height):
                return i
        return -1

    def _snap_position(self, idx: int, new_x: float, new_y: float) -> tuple[int, int]:
        """Snap monitor position to edges of other monitors."""
        m = self._monitors[idx]
        w = m.logical_width
        h = m.logical_height
        threshold = SNAP_DISTANCE / self._zoom

        best_x = new_x
        best_y = new_y
        min_dx = threshold + 1
        min_dy = threshold + 1

        edges_x = [new_x, new_x + w]
        edges_y = [new_y, new_y + h]

        for i, other in enumerate(self._monitors):
            if i == idx:
                continue
            ow = other.logical_width
            oh = other.logical_height

            # Target edges for X snapping
            targets_x = [other.x, other.x + ow]
            for ex in edges_x:
                for tx in targets_x:
                    d = abs(ex - tx)
                    if d < min_dx:
                        min_dx = d
                        best_x = new_x + (tx - ex)

            # Target edges for Y snapping
            targets_y = [other.y, other.y + oh]
            for ey in edges_y:
                for ty in targets_y:
                    d = abs(ey - ty)
                    if d < min_dy:
                        min_dy = d
                        best_y = new_y + (ty - ey)

        snap_x = best_x if min_dx <= threshold else new_x
        snap_y = best_y if min_dy <= threshold else new_y

        return round(snap_x), round(snap_y)

    # ── Event handlers ───────────────────────────────────────────────

    def _on_click_pressed(self, gesture: Gtk.GestureClick, n_press: int, x: float, y: float) -> None:
        idx = self._hit_test(x, y)
        if idx >= 0:
            self.selected_index = idx
            if n_press == 2:
                # Cancel any active drag and restore original position
                if self._dragging:
                    m = self._monitors[self._selected]
                    m.x = self._drag_orig_mx
                    m.y = self._drag_orig_my
                    self._dragging = False
                    self.queue_draw()
                self.emit("monitor-double-clicked", idx)

    def _on_mid_pressed(self, gesture: Gtk.GestureClick, n_press: int, x: float, y: float) -> None:
        pass  # handled by pan drag

    def _on_drag_begin(self, gesture: Gtk.GestureDrag, x: float, y: float) -> None:
        idx = self._hit_test(x, y)
        if idx >= 0:
            m = self._monitors[idx]
            self._dragging = True
            self._drag_start_x = x
            self._drag_start_y = y
            self._drag_orig_mx = m.x
            self._drag_orig_my = m.y
        else:
            self._dragging = False

    def _on_drag_update(self, gesture: Gtk.GestureDrag, offset_x: float, offset_y: float) -> None:
        if not self._dragging or self._selected < 0:
            return
        # Ignore micro-movements (likely double-click)
        if abs(offset_x) < 3 and abs(offset_y) < 3:
            return
        m = self._monitors[self._selected]
        new_x = self._drag_orig_mx + offset_x / self._zoom
        new_y = self._drag_orig_my + offset_y / self._zoom
        m.x, m.y = self._snap_position(self._selected, new_x, new_y)
        m.position_mode = PositionMode.EXPLICIT
        self.queue_draw()

    def _on_drag_end(self, gesture: Gtk.GestureDrag, offset_x: float, offset_y: float) -> None:
        if self._dragging and self._selected >= 0:
            m = self._monitors[self._selected]
            actually_moved = (m.x != self._drag_orig_mx or m.y != self._drag_orig_my)
            self._dragging = False
            if actually_moved:
                self.emit("monitor-moved", self._selected)

    def _on_pan_begin(self, gesture: Gtk.GestureDrag, x: float, y: float) -> None:
        self._panning = True
        self._pan_start_x = x
        self._pan_start_y = y
        self._pan_orig_px = self._pan_x
        self._pan_orig_py = self._pan_y

    def _on_pan_update(self, gesture: Gtk.GestureDrag, offset_x: float, offset_y: float) -> None:
        if self._panning:
            self._pan_x = self._pan_orig_px + offset_x
            self._pan_y = self._pan_orig_py + offset_y
            self.queue_draw()

    def _on_scroll(self, controller: Gtk.EventControllerScroll, dx: float, dy: float) -> bool:
        factor = 1.15 if dy < 0 else 1 / 1.15
        new_zoom = self._zoom * factor
        new_zoom = max(MIN_ZOOM, min(MAX_ZOOM, new_zoom))

        # Zoom toward cursor position
        # Get the current cursor position from the allocation center as fallback
        w = self.get_width()
        h = self.get_height()
        cx, cy = w / 2, h / 2

        # Adjust pan to zoom toward center
        self._pan_x = cx - (cx - self._pan_x) * (new_zoom / self._zoom)
        self._pan_y = cy - (cy - self._pan_y) * (new_zoom / self._zoom)
        self._zoom = new_zoom
        self.queue_draw()
        return True

    # ── Drawing ──────────────────────────────────────────────────────

    def _draw(self, area: Gtk.DrawingArea, cr, width: int, height: int) -> None:
        # Background
        cr.set_source_rgb(*COLOR_BG)
        cr.paint()

        # Grid
        self._draw_grid(cr, width, height)

        # Monitors
        for i, m in enumerate(self._monitors):
            self._draw_monitor(cr, m, i == self._selected)

    def _draw_grid(self, cr, width: int, height: int) -> None:
        cr.set_source_rgb(*COLOR_GRID)
        cr.set_line_width(0.5)

        # Determine grid spacing in logical coords
        base_spacing = 500  # logical pixels
        spacing = base_spacing * self._zoom

        # Ensure grid isn't too dense or too sparse
        while spacing < 30:
            spacing *= 2
        while spacing > 200:
            spacing /= 2

        # Draw vertical lines
        start_x = self._pan_x % spacing
        x = start_x
        while x < width:
            cr.move_to(x, 0)
            cr.line_to(x, height)
            x += spacing

        # Draw horizontal lines
        start_y = self._pan_y % spacing
        y = start_y
        while y < height:
            cr.move_to(0, y)
            cr.line_to(width, y)
            y += spacing

        cr.stroke()

    def _draw_monitor(self, cr, m: MonitorConfig, selected: bool) -> None:
        sx, sy = self._logical_to_screen(m.x, m.y)
        sw = m.logical_width * self._zoom
        sh = m.logical_height * self._zoom

        # Fill
        if not m.enabled:
            cr.set_source_rgb(*COLOR_DISABLED)
        elif selected:
            cr.set_source_rgb(*COLOR_SELECTED_FILL)
        else:
            cr.set_source_rgb(*COLOR_MONITOR)

        _rounded_rect(cr, sx, sy, sw, sh, 4)
        cr.fill()

        # Border
        if selected:
            cr.set_source_rgb(*COLOR_SELECTED)
            cr.set_line_width(2.5)
        else:
            cr.set_source_rgb(*COLOR_MONITOR_BORDER)
            cr.set_line_width(1.0)

        _rounded_rect(cr, sx, sy, sw, sh, 4)
        cr.stroke()

        # Text (only if monitor is big enough)
        if sw > 40 and sh > 20:
            self._draw_monitor_text(cr, m, sx, sy, sw, sh)

    def _draw_monitor_text(self, cr, m: MonitorConfig, sx: float, sy: float, sw: float, sh: float) -> None:
        # Clip text to monitor rectangle
        cr.save()
        padding = 4
        cr.rectangle(sx + padding, sy + padding, sw - padding * 2, sh - padding * 2)
        cr.clip()

        cr.set_source_rgb(*COLOR_TEXT)

        # Name — show description when enabled, fall back to port name
        name = (m.description if self._use_description and m.description else m.name) or "?"
        font_size = min(14, max(8, sw / 10))
        cr.set_font_size(font_size)

        extents = cr.text_extents(name)
        tx = sx + (sw - extents.width) / 2
        ty = sy + sh / 2 - 2
        cr.move_to(tx, ty)
        cr.show_text(name)

        # Resolution
        cr.set_source_rgb(*COLOR_TEXT_DIM)
        res_text = f"{m.width}x{m.height}"
        font_size_small = min(10, max(6, sw / 14))
        cr.set_font_size(font_size_small)
        extents = cr.text_extents(res_text)
        tx = sx + (sw - extents.width) / 2
        ty = sy + sh / 2 + font_size_small + 4
        if ty + 4 < sy + sh:
            cr.move_to(tx, ty)
            cr.show_text(res_text)

        cr.restore()


def _rounded_rect(cr, x: float, y: float, w: float, h: float, r: float) -> None:
    """Draw a rounded rectangle path."""
    cr.new_sub_path()
    cr.arc(x + w - r, y + r, r, -math.pi / 2, 0)
    cr.arc(x + w - r, y + h - r, r, 0, math.pi / 2)
    cr.arc(x + r, y + h - r, r, math.pi / 2, math.pi)
    cr.arc(x + r, y + r, r, math.pi, 3 * math.pi / 2)
    cr.close_path()



"""Sway IPC communication via binary i3-ipc protocol."""

from __future__ import annotations

import asyncio
import json
import os
import socket
import struct
from typing import AsyncIterator

from .models import MonitorConfig, Profile
from .utils import (
    is_hyprland_installed,
    hyprland_config_dir,
    sway_config_dir,
    is_sddm_running,
    is_greetd_running,
    write_xsetup,
    write_greetd_monitors,
    write_text,
    backup_file,
)

# i3-ipc protocol constants
_MAGIC = b"i3-ipc"
_HEADER_SIZE = 14  # 6 (magic) + 4 (payload_len) + 4 (type)
_HEADER_FMT = f"={len(_MAGIC)}sII"

# Message types
IPC_COMMAND = 0
IPC_GET_WORKSPACES = 1
IPC_SUBSCRIBE = 2
IPC_GET_OUTPUTS = 3

# Event type (high bit set)
EVENT_OUTPUT = 0x80000001


def _recv_exactly(sock: socket.socket, n: int) -> bytes:
    """Read exactly n bytes from a socket."""
    buf = bytearray()
    while len(buf) < n:
        chunk = sock.recv(n - len(buf))
        if not chunk:
            raise ConnectionError("Socket closed while reading")
        buf.extend(chunk)
    return bytes(buf)


async def _arecv_exactly(reader: asyncio.StreamReader, n: int) -> bytes:
    """Read exactly n bytes from an async StreamReader."""
    data = await reader.readexactly(n)
    return data


class SwayIPC:
    """Communicate with Sway via the i3-ipc binary protocol."""

    def __init__(self) -> None:
        self._socket_path = os.environ.get("SWAYSOCK", "")

    def _send(self, msg_type: int, payload: str = "") -> dict | list:
        """Send a message and return the parsed JSON response."""
        payload_bytes = payload.encode()
        header = struct.pack(_HEADER_FMT, _MAGIC, len(payload_bytes), msg_type)

        sock = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
        try:
            sock.connect(self._socket_path)
            sock.sendall(header + payload_bytes)

            # Read response header
            resp_header = _recv_exactly(sock, _HEADER_SIZE)
            _, resp_len, _ = struct.unpack(_HEADER_FMT, resp_header)

            # Read response payload
            resp_payload = _recv_exactly(sock, resp_len)
            return json.loads(resp_payload.decode())
        finally:
            sock.close()

    def get_outputs(self) -> list[dict]:
        """Query all outputs via GET_OUTPUTS."""
        return self._send(IPC_GET_OUTPUTS)

    def get_monitors(self) -> list[MonitorConfig]:
        """Query all connected monitors as MonitorConfig list."""
        outputs = self.get_outputs()
        return [MonitorConfig.from_sway_output(o) for o in outputs]

    def get_workspaces(self) -> list[dict]:
        """Query active workspaces."""
        return self._send(IPC_GET_WORKSPACES)

    def move_workspace_to_monitor(self, workspace: str, monitor: str) -> dict | list:
        """Move a workspace to a different monitor."""
        return self._send(IPC_COMMAND, f"[workspace={workspace}] move workspace to output {monitor}")

    def reload(self) -> dict:
        """Reload Sway configuration."""
        return self._send(IPC_COMMAND, "reload")

    def apply_profile(
        self, profile: Profile, *, update_sddm: bool = True,
        update_greetd: bool = True, use_description: bool = False,
    ) -> None:
        """Write monitor config and reload Sway."""
        conf_dir = sway_config_dir()
        monitors_conf = conf_dir / "monitors.conf"

        # Backup existing
        backup_file(monitors_conf)

        # Write Sway config
        write_text(monitors_conf, profile.generate_sway_config(use_description=use_description))

        # Also write Hyprland config if Hyprland is installed
        if is_hyprland_installed():
            hypr_conf = hyprland_config_dir() / "monitors.conf"
            backup_file(hypr_conf)
            write_text(hypr_conf, profile.generate_config(use_description=use_description))

        # Write SDDM Xsetup script if enabled and SDDM is present
        if update_sddm and is_sddm_running():
            write_xsetup(profile.generate_xsetup_script())

        # Write greetd sway monitors config if enabled and greetd is present
        if update_greetd and is_greetd_running():
            write_greetd_monitors(profile.generate_sway_config(use_description=use_description))

        # Reload
        self.reload()

    async def connect_event_socket(self) -> AsyncIterator[dict]:
        """Subscribe to output events and yield only output event payloads."""
        reader, writer = await asyncio.open_unix_connection(self._socket_path)

        try:
            # Subscribe to output events
            subscribe_payload = json.dumps(["output"]).encode()
            header = struct.pack(
                _HEADER_FMT, _MAGIC, len(subscribe_payload), IPC_SUBSCRIBE,
            )
            writer.write(header + subscribe_payload)
            await writer.drain()

            # Consume the subscribe ACK
            ack_header = await _arecv_exactly(reader, _HEADER_SIZE)
            _, ack_len, _ = struct.unpack(_HEADER_FMT, ack_header)
            await _arecv_exactly(reader, ack_len)

            # Yield only output events
            while True:
                evt_header = await _arecv_exactly(reader, _HEADER_SIZE)
                _, evt_len, evt_type = struct.unpack(_HEADER_FMT, evt_header)
                evt_payload = await _arecv_exactly(reader, evt_len)

                if evt_type == EVENT_OUTPUT:
                    yield json.loads(evt_payload.decode())
        finally:
            writer.close()



"""Background daemon that listens for monitor hotplug events and applies profiles."""

from __future__ import annotations

import asyncio
import logging
import os
import signal
import threading

from .hyprland import HyprlandIPC
from .sway import SwayIPC
from .models import Profile, apply_clamshell, undo_clamshell
from .profile_manager import ProfileManager
from .utils import load_app_settings

logging.basicConfig(
    level=logging.INFO,
    format="%(asctime)s [moniqued] %(levelname)s %(message)s",
)
log = logging.getLogger(__name__)

DEBOUNCE_MS = 500


def _detect_backend() -> HyprlandIPC | SwayIPC | None:
    """Auto-detect the running compositor from environment variables."""
    if os.environ.get("HYPRLAND_INSTANCE_SIGNATURE"):
        return HyprlandIPC()
    if os.environ.get("SWAYSOCK"):
        return SwayIPC()
    return None


class MonitorDaemon:
    """Watches compositor events and auto-applies matching profiles."""

    def __init__(self) -> None:
        self._profile_mgr = ProfileManager()
        self._debounce_handle: asyncio.TimerHandle | None = None
        self._ipc: HyprlandIPC | SwayIPC | None = None
        self._lid_closed: bool | None = None  # None = no lid / not monitored
        self._asyncio_loop: asyncio.AbstractEventLoop | None = None

    async def run(self) -> None:
        log.info("Starting Monique daemon")
        self._asyncio_loop = asyncio.get_event_loop()
        self._start_lid_monitor()

        while True:
            try:
                ipc = _detect_backend()
                if ipc is None:
                    log.warning("No supported compositor detected. Retrying in 5s...")
                    await asyncio.sleep(5)
                    continue

                backend_name = "Hyprland" if isinstance(ipc, HyprlandIPC) else "Sway"
                log.info("Detected %s compositor", backend_name)
                await self._listen(ipc)
            except (ConnectionRefusedError, FileNotFoundError, ConnectionError) as e:
                log.warning("Cannot connect to compositor: %s. Retrying in 5s...", e)
                await asyncio.sleep(5)
            except Exception as e:
                log.error("Unexpected error: %s. Retrying in 5s...", e)
                await asyncio.sleep(5)
            finally:
                if self._debounce_handle:
                    self._debounce_handle.cancel()
                    self._debounce_handle = None

    async def _listen(self, ipc: HyprlandIPC | SwayIPC) -> None:
        self._ipc = ipc
        log.info("Connected to compositor event socket")
        async for event in ipc.connect_event_socket():
            log.info("Monitor event: %s", event)
            self._schedule_apply(ipc)

    def _schedule_apply(self, ipc: HyprlandIPC | SwayIPC) -> None:
        """Debounce monitor events before applying."""
        loop = asyncio.get_event_loop()
        if self._debounce_handle:
            self._debounce_handle.cancel()
        self._debounce_handle = loop.call_later(
            DEBOUNCE_MS / 1000.0,
            lambda: asyncio.ensure_future(self._apply_best_profile(ipc)),
        )

    async def _apply_best_profile(self, ipc: HyprlandIPC | SwayIPC) -> None:
        """Query current monitors, find best profile, and apply it."""
        try:
            monitors = ipc.get_monitors()
            fingerprint = sorted(m.description for m in monitors if m.description)
            log.info("Current fingerprint: %s", fingerprint)

            settings = load_app_settings()
            clamshell = settings.get("clamshell_mode", False)

            profile = self._profile_mgr.find_best_match(fingerprint)
            if profile:
                # Apply clamshell mode if enabled
                if clamshell:
                    profile = Profile.from_dict(profile.to_dict())
                    if self._lid_closed is not False:
                        # Lid closed or unknown → disable internal
                        if apply_clamshell(profile.monitors):
                            log.info("Clamshell: disabled internal display(s)")
                    else:
                        # Lid open → ensure internal is enabled
                        if undo_clamshell(profile.monitors):
                            log.info("Clamshell: lid open, re-enabled internal display(s)")

                # Snapshot workspaces before applying
                ws_snapshot = ipc.get_workspaces()

                log.info("Applying profile: %s", profile.name)
                update_sddm = settings.get("update_sddm", True)
                update_greetd = settings.get("update_greetd", True)
                use_desc = not settings.get("use_port_names", False)
                ipc.apply_profile(
                    profile, update_sddm=update_sddm,
                    update_greetd=update_greetd, use_description=use_desc,
                )

                # Migrate orphaned workspaces
                if settings.get("migrate_workspaces", True):
                    self._migrate_orphaned_workspaces(ipc, profile, ws_snapshot)
            elif clamshell and self._lid_closed is False:
                # No saved profile, lid open → re-enable internal
                if undo_clamshell(monitors):
                    log.info("Clamshell: re-enabled internal display(s)")
                    temp = Profile(name="clamshell-undo", monitors=monitors)
                    update_sddm = settings.get("update_sddm", True)
                    use_desc = not settings.get("use_port_names", False)
                    ipc.apply_profile(
                        temp, update_sddm=update_sddm, use_description=use_desc,
                    )
            else:
                log.info("No matching profile found")
        except Exception as e:
            log.error("Failed to apply profile: %s", e)

    # ── Lid monitoring via UPower D-Bus ─────────────────────────────

    def _start_lid_monitor(self) -> None:
        """Start monitoring lid state via UPower D-Bus in a background thread."""
        try:
            import gi
            gi.require_version("Gio", "2.0")
            from gi.repository import Gio, GLib
        except (ImportError, ValueError):
            log.info("GLib not available, lid monitoring disabled")
            return

        def _run() -> None:
            try:
                bus = Gio.bus_get_sync(Gio.BusType.SYSTEM)

                # Check if lid is present
                result = bus.call_sync(
                    "org.freedesktop.UPower",
                    "/org/freedesktop/UPower",
                    "org.freedesktop.DBus.Properties",
                    "Get",
                    GLib.Variant("(ss)", ("org.freedesktop.UPower", "LidIsPresent")),
                    GLib.VariantType("(v)"),
                    Gio.DBusCallFlags.NONE,
                    -1,
                    None,
                )
                lid_present = result.get_child_value(0).get_variant().get_boolean()
                if not lid_present:
                    log.info("No lid detected, lid monitoring disabled")
                    return

                # Get initial lid state
                result = bus.call_sync(
                    "org.freedesktop.UPower",
                    "/org/freedesktop/UPower",
                    "org.freedesktop.DBus.Properties",
                    "Get",
                    GLib.Variant("(ss)", ("org.freedesktop.UPower", "LidIsClosed")),
                    GLib.VariantType("(v)"),
                    Gio.DBusCallFlags.NONE,
                    -1,
                    None,
                )
                self._lid_closed = result.get_child_value(0).get_variant().get_boolean()
                log.info("Initial lid state: %s", "closed" if self._lid_closed else "open")

                # Subscribe to PropertiesChanged
                bus.signal_subscribe(
                    "org.freedesktop.UPower",
                    "org.freedesktop.DBus.Properties",
                    "PropertiesChanged",
                    "/org/freedesktop/UPower",
                    None,
                    Gio.DBusSignalFlags.NONE,
                    _on_signal,
                    None,
                )

                loop = GLib.MainLoop.new(GLib.MainContext.default(), False)
                loop.run()
            except Exception as e:
                log.warning("Lid monitor failed: %s", e)

        def _on_signal(_conn, _sender, _path, _iface, _signal, params, _user_data):
            iface_name = params.get_child_value(0).get_string()
            if iface_name != "org.freedesktop.UPower":
                return
            changed = params.get_child_value(1)
            lid_val = changed.lookup_value("LidIsClosed", GLib.VariantType("b"))
            if lid_val is None:
                return
            closed = lid_val.get_boolean()
            log.info("Lid state changed: %s", "closed" if closed else "open")
            self._lid_closed = closed
            if self._ipc and self._asyncio_loop:
                self._asyncio_loop.call_soon_threadsafe(
                    self._schedule_apply, self._ipc,
                )

        thread = threading.Thread(target=_run, daemon=True, name="lid-monitor")
        thread.start()

    def _migrate_orphaned_workspaces(
        self,
        ipc: HyprlandIPC | SwayIPC,
        profile,
        ws_snapshot: list[dict],
    ) -> None:
        """Move workspaces from disabled/removed monitors to the primary monitor."""
        enabled_names = {m.name for m in profile.monitors if m.enabled}
        if not enabled_names:
            return

        primary = next(m.name for m in profile.monitors if m.enabled)
        migrated = 0

        for ws in ws_snapshot:
            ws_monitor = ws.get("monitor", "")
            ws_name = str(ws.get("name", ws.get("id", "")))
            if ws_monitor and ws_monitor not in enabled_names:
                try:
                    ipc.move_workspace_to_monitor(ws_name, primary)
                    migrated += 1
                except Exception as e:
                    log.warning("Failed to migrate workspace %s: %s", ws_name, e)

        if migrated:
            log.info("Migrated %d workspace(s) to %s", migrated, primary)


def main() -> None:
    daemon = MonitorDaemon()
    loop = asyncio.new_event_loop()

    # Handle signals for clean shutdown
    for sig in (signal.SIGTERM, signal.SIGINT):
        loop.add_signal_handler(sig, loop.stop)

    try:
        loop.run_until_complete(daemon.run())
    except KeyboardInterrupt:
        pass
    finally:
        loop.close()
        log.info("Daemon stopped")



"""Workspace assignment dialog."""

from __future__ import annotations

import gi
gi.require_version("Gtk", "4.0")
gi.require_version("Adw", "1")
from gi.repository import Gtk, Gdk, Adw, GLib, GObject

from .models import WorkspaceRule


class WorkspacePanel(Adw.Window):
    """Window for managing workspace-to-monitor assignments."""

    __gtype_name__ = "WorkspacePanel"

    __gsignals__ = {
        "rules-changed": (GObject.SignalFlags.RUN_FIRST, None, ()),
    }

    def __init__(
        self,
        monitor_names: list[str] | None = None,
        monitor_descriptions: list[str] | None = None,
        monitor_enabled: list[bool] | None = None,
        application: Gtk.Application | None = None,
    ) -> None:
        super().__init__(application=application)
        self._rules: list[WorkspaceRule] = []
        self._monitor_names = monitor_names or []
        self._monitor_descriptions = monitor_descriptions or []
        self._drag_index: int = -1
        self._monitor_enabled = monitor_enabled or [True] * len(self._monitor_names)
        self._building = False

        self.set_title("Workspace Rules")
        self.set_modal(True)
        # Size to ~45% width, ~70% height of the largest monitor,
        # but never smaller than 660x800 (unless the screen is smaller)
        w, h = 660, 800
        display = Gdk.Display.get_default()
        if display:
            mons = display.get_monitors()
            max_w, max_h = 0, 0
            for i in range(mons.get_n_items()):
                geom = mons.get_item(i).get_geometry()
                max_w = max(max_w, geom.width)
                max_h = max(max_h, geom.height)
            if max_w > 0:
                w = min(max_w, max(w, int(max_w * 0.45)))
                h = min(max_h, max(h, int(max_h * 0.7)))
        self.set_default_size(w, h)

        self._build_ui()

    # -- Helpers --------------------------------------------------------------

    @staticmethod
    def _fix_spin_icons(widget: Gtk.Widget) -> None:
        """Replace SpinRow +/- button icons with text labels after a delay."""
        _LABELS = {
            "list-add-symbolic": "+",
            "list-remove-symbolic": "\u2212",
            "value-increase-symbolic": "+",
            "value-decrease-symbolic": "\u2212",
        }

        def _do_fix():
            _walk(widget)
            return False  # don't repeat

        def _walk(w):
            child = w.get_first_child()
            while child:
                next_s = child.get_next_sibling()
                # Check if this is a button with a spin icon
                if isinstance(child, Gtk.Button):
                    icon = child.get_icon_name() or ""
                    if not icon:
                        bc = child.get_child()
                        if isinstance(bc, Gtk.Image):
                            icon = bc.get_icon_name() or ""
                    if icon in _LABELS:
                        child.set_icon_name("")
                        lbl = Gtk.Label(label=_LABELS[icon])
                        lbl.add_css_class("heading")
                        child.set_child(lbl)
                # Always recurse into all children
                _walk(child)
                child = next_s

        GLib.timeout_add(150, _do_fix)

    # -- UI building ----------------------------------------------------------

    def _build_ui(self) -> None:
        box = Gtk.Box(orientation=Gtk.Orientation.VERTICAL)

        # Header bar
        header = Adw.HeaderBar()
        header.set_show_end_title_buttons(True)
        header.set_show_start_title_buttons(False)

        btn_add = Gtk.Button()
        btn_add.set_child(Gtk.Image.new_from_icon_name("list-add-symbolic"))
        btn_add.set_tooltip_text("Add rule")
        btn_add.connect("clicked", self._on_add_clicked)
        header.pack_start(btn_add)

        btn_quick = Gtk.Button()
        btn_quick.set_child(Gtk.Image.new_from_icon_name("document-new-symbolic"))
        btn_quick.set_tooltip_text("Quick Setup")
        btn_quick.connect("clicked", self._on_quick_setup_clicked)
        header.pack_start(btn_quick)

        box.append(header)

        # Stack to switch between empty state and rules list
        self._stack = Gtk.Stack()
        self._stack.set_vexpand(True)
        self._stack.set_transition_type(Gtk.StackTransitionType.CROSSFADE)

        # Empty status page
        self._status_page = Adw.StatusPage()
        self._status_page.set_icon_name("view-grid-symbolic")
        self._status_page.set_title("No Workspace Rules")
        self._status_page.set_description(
            "Configure workspace assignments to monitors"
        )
        btn_setup = Gtk.Button(label="Quick Setup")
        btn_setup.add_css_class("suggested-action")
        btn_setup.add_css_class("pill")
        btn_setup.set_halign(Gtk.Align.CENTER)
        btn_setup.connect("clicked", self._on_quick_setup_clicked)
        self._status_page.set_child(btn_setup)
        self._stack.add_named(self._status_page, "empty")

        # Scrolled content with rules list
        scroll = Gtk.ScrolledWindow(vexpand=True)
        self._list_box = Gtk.ListBox()
        self._list_box.set_selection_mode(Gtk.SelectionMode.NONE)
        self._list_box.add_css_class("boxed-list")
        self._list_box.set_margin_start(12)
        self._list_box.set_margin_end(12)
        self._list_box.set_margin_top(12)
        self._list_box.set_margin_bottom(12)
        scroll.set_child(self._list_box)
        self._stack.add_named(scroll, "rules")

        box.append(self._stack)
        self.set_content(box)

        self._update_stack_visible()

    def _update_stack_visible(self) -> None:
        if self._rules:
            self._stack.set_visible_child_name("rules")
        else:
            self._stack.set_visible_child_name("empty")

    # -- Rules list -----------------------------------------------------------

    def set_rules(self, rules: list[WorkspaceRule]) -> None:
        """Populate the dialog with existing rules."""
        self._rules = [WorkspaceRule(**r.__dict__) for r in rules]  # copy
        self._rebuild_list()

    def get_rules(self) -> list[WorkspaceRule]:
        """Return current rules."""
        return list(self._rules)

    def _rebuild_list(self) -> None:
        self._building = True
        while True:
            row = self._list_box.get_row_at_index(0)
            if row is None:
                break
            self._list_box.remove(row)

        for i, rule in enumerate(self._rules):
            row = self._create_rule_row(i, rule)
            self._list_box.append(row)

        self._building = False
        self._update_stack_visible()

    # -- Monitor dropdown helpers ---------------------------------------------

    def _monitor_full_options(self) -> list[str]:
        """Full description strings for popup list."""
        options = ["None"]
        for i, name in enumerate(self._monitor_names):
            desc = (
                self._monitor_descriptions[i]
                if i < len(self._monitor_descriptions)
                   and self._monitor_descriptions[i]
                else ""
            )
            options.append(f"{desc} ({name})" if desc else name)
        return options

    def _monitor_compact_options(self) -> list[str]:
        """Compact strings for button display."""
        options = ["None"]
        for i, name in enumerate(self._monitor_names):
            desc = (
                self._monitor_descriptions[i]
                if i < len(self._monitor_descriptions)
                   and self._monitor_descriptions[i]
                else ""
            )
            if desc:
                short = desc[:10].rstrip()
                ellipsis = "\u2026" if len(desc) > 10 else ""
                options.append(f"{short}{ellipsis} ({name})")
            else:
                options.append(name)
        return options

    def _make_monitor_combo(self) -> Gtk.DropDown:
        """Build a DropDown with compact button and full-description popup."""
        full = self._monitor_full_options()
        compact = self._monitor_compact_options()
        model = Gtk.StringList.new(full)

        # Factory for the button area (compact)
        btn_factory = Gtk.SignalListItemFactory()

        def setup_btn(factory, list_item):
            lbl = Gtk.Label(xalign=0, ellipsize=2)  # END
            list_item.set_child(lbl)

        def bind_btn(factory, list_item):
            lbl = list_item.get_child()
            pos = list_item.get_position()
            lbl.set_label(compact[pos] if pos < len(compact) else "")

        btn_factory.connect("setup", setup_btn)
        btn_factory.connect("bind", bind_btn)

        # Factory for the popup list (full description)
        list_factory = Gtk.SignalListItemFactory()

        def setup_list(factory, list_item):
            lbl = Gtk.Label(xalign=0)
            list_item.set_child(lbl)

        def bind_list(factory, list_item):
            lbl = list_item.get_child()
            lbl.set_label(list_item.get_item().get_string())

        list_factory.connect("setup", setup_list)
        list_factory.connect("bind", bind_list)

        combo = Gtk.DropDown(model=model)
        combo.set_factory(btn_factory)
        combo.set_list_factory(list_factory)
        return combo

    # -- Rule row -------------------------------------------------------------

    def _create_rule_row(self, index: int, rule: WorkspaceRule) -> Adw.ExpanderRow:
        expander = Adw.ExpanderRow()

        # Row counter (fixed, 1-based)
        counter_label = Gtk.Label(
            label=str(index + 1),
            width_chars=3,
            xalign=0.5,
            valign=Gtk.Align.CENTER,
        )
        counter_label.add_css_class("dim-label")
        expander.add_prefix(counter_label)

        # Suffixes are rendered right-to-left: first added = rightmost.
        # Visual order: [dropdown] [star] [pin] [delete]

        # Delete button (rightmost)
        btn_del = Gtk.Button()
        btn_del.set_child(Gtk.Image.new_from_icon_name("edit-delete-symbolic"))
        btn_del.set_valign(Gtk.Align.CENTER)
        btn_del.add_css_class("destructive-action")
        btn_del.add_css_class("flat")
        btn_del.connect("clicked", self._on_delete_clicked, index)
        expander.add_suffix(btn_del)

        # Persistent toggle (pin)
        btn_persistent = Gtk.ToggleButton()
        btn_persistent.set_child(Gtk.Image.new_from_icon_name("view-pin-symbolic"))
        btn_persistent.set_active(rule.persistent)
        btn_persistent.set_valign(Gtk.Align.CENTER)
        btn_persistent.add_css_class("flat")
        if rule.persistent:
            btn_persistent.add_css_class("accent")
        btn_persistent.set_tooltip_text("Persistent workspace")
        btn_persistent.connect("toggled", self._on_persistent_toggled, index)
        expander.add_suffix(btn_persistent)

        # Default toggle (star)
        btn_default = Gtk.ToggleButton()
        btn_default.set_child(Gtk.Image.new_from_icon_name("starred-symbolic"))
        btn_default.set_active(rule.default)
        btn_default.set_valign(Gtk.Align.CENTER)
        btn_default.add_css_class("flat")
        if rule.default:
            btn_default.add_css_class("warning")
        btn_default.set_tooltip_text("Default workspace for this monitor")
        btn_default.connect("toggled", self._on_default_toggled, index)
        expander.add_suffix(btn_default)

        # Monitor combo (leftmost suffix)
        monitor_combo = self._make_monitor_combo()
        if rule.monitor in self._monitor_names:
            monitor_combo.set_selected(self._monitor_names.index(rule.monitor) + 1)
        else:
            monitor_combo.set_selected(0)
        monitor_combo.set_valign(Gtk.Align.CENTER)
        monitor_combo.set_size_request(240, -1)
        monitor_combo.connect("notify::selected", self._on_monitor_changed, index)
        expander.add_suffix(monitor_combo)

        # Title: "Workspace N" if numeric, just the name if custom
        expander.set_title(self._ws_title(rule.workspace))

        # -- Rows inside expander --

        # Workspace entry for editing
        entry_ws = Adw.EntryRow(title="Workspace name / number")
        entry_ws.set_text(rule.workspace)
        entry_ws.connect("changed", self._on_ws_changed, index, expander)
        expander.add_row(entry_ws)

        spin_rounding = Adw.SpinRow.new_with_range(-1, 1, 1)
        spin_rounding.set_title("Rounding (-1=unset)")
        spin_rounding.add_prefix(Gtk.Image.new_from_icon_name("draw-circle-symbolic"))
        spin_rounding.set_value(rule.rounding)
        spin_rounding.connect("notify::value", self._on_spin_changed, index, "rounding")
        expander.add_row(spin_rounding)
        self._fix_spin_icons(spin_rounding)

        spin_decorate = Adw.SpinRow.new_with_range(-1, 1, 1)
        spin_decorate.set_title("Decorate (-1=unset)")
        spin_decorate.add_prefix(Gtk.Image.new_from_icon_name("applications-graphics-symbolic"))
        spin_decorate.set_value(rule.decorate)
        spin_decorate.connect("notify::value", self._on_spin_changed, index, "decorate")
        expander.add_row(spin_decorate)
        self._fix_spin_icons(spin_decorate)

        spin_gapsin = Adw.SpinRow.new_with_range(-1, 100, 1)
        spin_gapsin.set_title("Gaps In (-1=unset)")
        spin_gapsin.add_prefix(Gtk.Image.new_from_icon_name("view-dual-symbolic"))
        spin_gapsin.set_value(rule.gapsin)
        spin_gapsin.connect("notify::value", self._on_spin_changed, index, "gapsin")
        expander.add_row(spin_gapsin)
        self._fix_spin_icons(spin_gapsin)

        spin_gapsout = Adw.SpinRow.new_with_range(-1, 100, 1)
        spin_gapsout.set_title("Gaps Out (-1=unset)")
        spin_gapsout.add_prefix(Gtk.Image.new_from_icon_name("view-dual-symbolic"))
        spin_gapsout.set_value(rule.gapsout)
        spin_gapsout.connect("notify::value", self._on_spin_changed, index, "gapsout")
        expander.add_row(spin_gapsout)
        self._fix_spin_icons(spin_gapsout)

        spin_border = Adw.SpinRow.new_with_range(-1, 1, 1)
        spin_border.set_title("Border (-1=unset)")
        spin_border.add_prefix(Gtk.Image.new_from_icon_name("focus-windows-symbolic"))
        spin_border.set_value(rule.border)
        spin_border.connect("notify::value", self._on_spin_changed, index, "border")
        expander.add_row(spin_border)
        self._fix_spin_icons(spin_border)

        spin_bordersize = Adw.SpinRow.new_with_range(-1, 20, 1)
        spin_bordersize.set_title("Border Size (-1=unset)")
        spin_bordersize.add_prefix(Gtk.Image.new_from_icon_name("value-increase-symbolic"))
        spin_bordersize.set_value(rule.bordersize)
        spin_bordersize.connect("notify::value", self._on_spin_changed, index, "bordersize")
        expander.add_row(spin_bordersize)
        self._fix_spin_icons(spin_bordersize)

        entry_oce = Adw.EntryRow(title="On Created Empty")
        entry_oce.add_prefix(Gtk.Image.new_from_icon_name("utilities-terminal-symbolic"))
        entry_oce.set_text(rule.on_created_empty)
        entry_oce.connect("changed", self._on_text_changed, index, "on_created_empty")
        expander.add_row(entry_oce)

        # Drag-and-drop reordering
        drag = Gtk.DragSource()
        drag.set_actions(Gdk.DragAction.MOVE)
        drag.connect("prepare", self._on_drag_prepare, index)
        expander.add_controller(drag)

        drop = Gtk.DropTarget.new(GObject.TYPE_STRING, Gdk.DragAction.MOVE)
        drop.connect("drop", self._on_drop, index)
        expander.add_controller(drop)

        return expander

    # -- Quick Setup ----------------------------------------------------------

    def _on_quick_setup_clicked(self, btn: Gtk.Button) -> None:
        if self._rules:
            self._confirm_replace_then_setup()
        else:
            self._show_quick_setup_dialog()

    def _confirm_replace_then_setup(self) -> None:
        dialog = Adw.AlertDialog()
        dialog.set_heading("Replace existing rules?")
        dialog.set_body(
            "Quick Setup will remove all current workspace rules "
            "and generate new ones."
        )
        dialog.add_response("cancel", "Cancel")
        dialog.add_response("replace", "Replace")
        dialog.set_response_appearance(
            "replace", Adw.ResponseAppearance.DESTRUCTIVE
        )
        dialog.set_default_response("cancel")
        dialog.set_close_response("cancel")
        dialog.connect("response", self._on_confirm_replace_response)
        dialog.present(self)

    def _on_confirm_replace_response(
        self, dialog: Adw.AlertDialog, response: str
    ) -> None:
        if response == "replace":
            self._show_quick_setup_dialog()

    def _show_quick_setup_dialog(self) -> None:
        dialog = Adw.Dialog()
        dialog.set_title("Quick Setup")
        dialog.set_content_width(500)
        dialog.set_content_height(550)

        outer = Gtk.Box(orientation=Gtk.Orientation.VERTICAL)

        header = Adw.HeaderBar()
        header.set_show_end_title_buttons(True)
        header.set_show_start_title_buttons(False)

        btn_generate = Gtk.Button(label="Generate")
        btn_generate.add_css_class("suggested-action")
        header.pack_end(btn_generate)
        outer.append(header)

        scroll = Gtk.ScrolledWindow(vexpand=True)
        page = Adw.PreferencesPage()

        # -- Total workspaces group --
        grp_total = Adw.PreferencesGroup(title="Workspaces")
        spin_total = Adw.SpinRow.new_with_range(1, 20, 1)
        spin_total.set_title("Number of workspaces")
        spin_total.set_value(10)
        grp_total.add(spin_total)
        self._fix_spin_icons(spin_total)
        page.add(grp_total)

        # -- Per-monitor distribution --
        # Sort: enabled first, disabled last
        monitor_order: list[int] = sorted(
            range(len(self._monitor_names)),
            key=lambda i: (not self._monitor_enabled[i], i),
        )
        # Parallel lists: spin values preserved across rebuilds
        spin_values: list[int] = [0] * len(self._monitor_names)

        grp_dist = Adw.PreferencesGroup(
            title="Distribution",
            description="Drag to reorder monitor priority",
        )
        monitor_listbox = Gtk.ListBox()
        monitor_listbox.add_css_class("boxed-list")
        monitor_listbox.set_selection_mode(Gtk.SelectionMode.NONE)
        grp_dist.add(monitor_listbox)

        summary_label = Gtk.Label(
            label="",
            halign=Gtk.Align.END,
            margin_top=8,
            margin_end=12,
            margin_bottom=4,
        )
        summary_label.add_css_class("dim-label")
        summary_label.add_css_class("caption")
        grp_dist.add(summary_label)

        page.add(grp_dist)

        scroll.set_child(page)
        outer.append(scroll)
        dialog.set_child(outer)

        # -- State helpers ---
        distributing = [False]
        qs_drag = [-1]  # mutable drag source rank
        spin_rows: list[Adw.SpinRow] = []

        def build_monitor_rows() -> None:
            """(Re)build the monitor ListBox from monitor_order."""
            # Save current values
            for i, spin in enumerate(spin_rows):
                spin_values[monitor_order[i]] = int(spin.get_value())

            spin_rows.clear()
            while True:
                row = monitor_listbox.get_row_at_index(0)
                if row is None:
                    break
                monitor_listbox.remove(row)

            for rank, idx in enumerate(monitor_order):
                name = self._monitor_names[idx]
                enabled = self._monitor_enabled[idx]
                spin = Adw.SpinRow.new_with_range(0, 20, 1)
                title = f"Workspaces for {name}"
                if not enabled:
                    title += " (disabled)"
                spin.set_title(title)
                if idx < len(self._monitor_descriptions) and self._monitor_descriptions[idx]:
                    spin.set_subtitle(self._monitor_descriptions[idx])
                    spin.set_subtitle_lines(1)
                    spin.set_title_lines(1)
                spin.set_margin_top(8)
                spin.set_margin_bottom(8)
                if not enabled:
                    spin.set_sensitive(False)
                spin.set_value(spin_values[idx])
                spin.connect("notify::value", on_monitor_spin_changed)

                # DnD
                drag = Gtk.DragSource()
                drag.set_actions(Gdk.DragAction.MOVE)
                drag.connect("prepare", on_qs_drag_prepare, rank)
                spin.add_controller(drag)

                drop = Gtk.DropTarget.new(GObject.TYPE_STRING, Gdk.DragAction.MOVE)
                drop.connect("drop", on_qs_drop, rank)
                spin.add_controller(drop)

                monitor_listbox.append(spin)
                spin_rows.append(spin)
                self._fix_spin_icons(spin)

        def on_qs_drag_prepare(source, x, y, rank):
            qs_drag[0] = rank
            return Gdk.ContentProvider.new_for_value(
                GObject.Value(GObject.TYPE_STRING, str(rank))
            )

        def on_qs_drop(target, value, x, y, dest_rank):
            src = qs_drag[0]
            if src < 0 or src == dest_rank or src >= len(monitor_order):
                return False
            idx = monitor_order.pop(src)
            monitor_order.insert(dest_rank, idx)
            build_monitor_rows()
            update_summary()
            return True

        def distribute(total: int) -> None:
            distributing[0] = True
            enabled = [r for r, i in enumerate(monitor_order) if self._monitor_enabled[i]]
            n = len(enabled)
            if n == 0:
                distributing[0] = False
                return
            base = total // n
            extra = total % n
            for pos, rank in enumerate(enabled):
                spin_rows[rank].set_value(base + (1 if pos < extra else 0))
            distributing[0] = False
            update_summary()

        def update_summary() -> None:
            assigned = sum(int(s.get_value()) for s in spin_rows)
            total = int(spin_total.get_value())
            summary_label.set_label(f"Assigned {assigned} / {total}")
            if assigned != total:
                summary_label.add_css_class("error")
            else:
                summary_label.remove_css_class("error")

        def on_total_changed(row: Adw.SpinRow, pspec) -> None:
            if not distributing[0]:
                distribute(int(row.get_value()))

        def on_monitor_spin_changed(row: Adw.SpinRow, pspec) -> None:
            if not distributing[0]:
                update_summary()

        spin_total.connect("notify::value", on_total_changed)

        # Build initial rows and distribute
        build_monitor_rows()
        distribute(int(spin_total.get_value()))

        def on_generate_clicked(btn: Gtk.Button) -> None:
            total = int(spin_total.get_value())
            # Use monitor_order for workspace assignment order
            distribution = {}
            for rank, idx in enumerate(monitor_order):
                distribution[self._monitor_names[idx]] = int(spin_rows[rank].get_value())
            dialog.force_close()
            self._generate_rules(total, distribution)

        btn_generate.connect("clicked", on_generate_clicked)
        dialog.present(self)

    def _generate_rules(
        self, total: int, distribution: dict[str, int]
    ) -> None:
        rules: list[WorkspaceRule] = []
        ws_num = 1
        for monitor, count in distribution.items():
            for i in range(count):
                rule = WorkspaceRule(
                    workspace=str(ws_num),
                    monitor=monitor,
                    default=(i == 0),
                    persistent=(i == 0),
                )
                rules.append(rule)
                ws_num += 1
        self._rules = rules
        self._rebuild_list()
        self.emit("rules-changed")

    # -- Handlers -------------------------------------------------------------

    def _on_add_clicked(self, btn: Gtk.Button) -> None:
        self._rules.append(WorkspaceRule())
        self._rebuild_list()
        self.emit("rules-changed")

    def _on_drag_prepare(self, source, x, y, index):
        self._drag_index = index
        return Gdk.ContentProvider.new_for_value(
            GObject.Value(GObject.TYPE_STRING, str(index))
        )

    def _on_drop(self, target, value, x, y, dest_index):
        src = self._drag_index
        if src < 0 or src == dest_index or src >= len(self._rules):
            return False
        rule = self._rules.pop(src)
        self._rules.insert(dest_index, rule)
        self._rebuild_list()
        self.emit("rules-changed")
        return True

    def _on_delete_clicked(self, btn: Gtk.Button, index: int) -> None:
        if 0 <= index < len(self._rules):
            self._rules.pop(index)
            self._rebuild_list()
            self.emit("rules-changed")

    @staticmethod
    def _ws_title(ws: str) -> str:
        if not ws:
            return "New Rule"
        if ws.isdigit():
            return f"Workspace {ws}"
        return ws

    def _on_ws_changed(self, entry, index: int, expander=None) -> None:
        if self._building or index >= len(self._rules):
            return
        text = entry.get_text()
        self._rules[index].workspace = text
        if expander:
            expander.set_title(self._ws_title(text))
        self.emit("rules-changed")

    def _on_monitor_changed(self, combo: Gtk.DropDown, pspec, index: int) -> None:
        if self._building or index >= len(self._rules):
            return
        sel = combo.get_selected()
        if sel == 0:
            self._rules[index].monitor = ""
        else:
            options = [""] + self._monitor_names
            if sel < len(options):
                self._rules[index].monitor = options[sel]
        self.emit("rules-changed")

    def _on_persistent_toggled(self, btn: Gtk.ToggleButton, index: int) -> None:
        if self._building or index >= len(self._rules):
            return
        active = btn.get_active()
        self._rules[index].persistent = active
        if active:
            btn.add_css_class("accent")
        else:
            btn.remove_css_class("accent")
        self.emit("rules-changed")

    def _on_default_toggled(self, btn: Gtk.ToggleButton, index: int) -> None:
        if self._building or index >= len(self._rules):
            return
        active = btn.get_active()
        self._rules[index].default = active
        # Ensure only one default per monitor
        if active:
            monitor = self._rules[index].monitor
            if monitor:
                for i, r in enumerate(self._rules):
                    if i != index and r.monitor == monitor and r.default:
                        r.default = False
            self._rebuild_list()
        else:
            btn.remove_css_class("warning")
        self.emit("rules-changed")

    def _on_advanced_changed(self, row: Adw.SwitchRow, pspec, index: int, field: str) -> None:
        if self._building or index >= len(self._rules):
            return
        setattr(self._rules[index], field, row.get_active())
        self.emit("rules-changed")

    def _on_spin_changed(self, row: Adw.SpinRow, pspec, index: int, field: str) -> None:
        if self._building or index >= len(self._rules):
            return
        setattr(self._rules[index], field, int(row.get_value()))
        self.emit("rules-changed")

    def _on_text_changed(self, row: Adw.EntryRow, index: int, field: str) -> None:
        if self._building or index >= len(self._rules):
            return
        setattr(self._rules[index], field, row.get_text())
        self.emit("rules-changed")



"""Main application window."""

from __future__ import annotations

import gi
gi.require_version("Gtk", "4.0")
gi.require_version("Adw", "1")
from gi.repository import Gtk, Gdk, Adw, GLib, Gio

from .canvas import MonitorCanvas
from .properties_panel import PropertiesPanel
from .workspace_panel import WorkspacePanel
from .profile_manager import ProfileManager
from .models import MonitorConfig, Profile, WorkspaceRule, apply_clamshell, undo_clamshell
from .hyprland import HyprlandIPC
from .utils import (
    hyprland_config_dir,
    backup_file,
    restore_backup,
    is_sddm_running,
    is_greetd_running,
    load_app_settings,
    save_app_settings,
)


CONFIRM_TIMEOUT = 10  # seconds
OSD_TIMEOUT_MS = 1500


OSD_WINDOW_TITLE = "monique-osd"


class MonitorOSD(Gtk.Window):
    """Fullscreen transparent overlay showing the monitor name."""

    _css_loaded = False
    _rules_set = False

    def __init__(self, app: Gtk.Application, gdk_monitor: Gdk.Monitor, label: str) -> None:
        super().__init__(application=app)
        self._timer_id: int = 0

        if not MonitorOSD._css_loaded:
            css = Gtk.CssProvider()
            css.load_from_string(
                ".osd-bg { background: rgba(0,0,0,0.35); }"
                ".osd-label { font-size: 72px; font-weight: 800; color: white;"
                " text-shadow: 0 2px 12px rgba(0,0,0,0.7); }"
                ".osd-sub { font-size: 24px; font-weight: 400; color: rgba(255,255,255,0.7);"
                " text-shadow: 0 1px 6px rgba(0,0,0,0.5); }"
            )
            Gtk.StyleContext.add_provider_for_display(
                Gdk.Display.get_default(), css, Gtk.STYLE_PROVIDER_PRIORITY_APPLICATION,
            )
            MonitorOSD._css_loaded = True

        self.set_title(OSD_WINDOW_TITLE)
        self.set_decorated(False)

        box = Gtk.Box(orientation=Gtk.Orientation.VERTICAL,
                      valign=Gtk.Align.CENTER, halign=Gtk.Align.CENTER)
        box.set_vexpand(True)
        box.set_hexpand(True)
        box.add_css_class("osd-bg")

        name_lbl = Gtk.Label(label=label)
        name_lbl.add_css_class("osd-label")
        box.append(name_lbl)

        geom = gdk_monitor.get_geometry()
        sub = Gtk.Label(label=f"{geom.width}x{geom.height}")
        sub.add_css_class("osd-sub")
        box.append(sub)

        self.set_child(box)
        self.fullscreen_on_monitor(gdk_monitor)

    def show_timed(self, timeout_ms: int = OSD_TIMEOUT_MS) -> None:
        self._ensure_rules()
        self.present()
        self._timer_id = GLib.timeout_add(timeout_ms, self._dismiss)

    def _ensure_rules(self) -> None:
        if MonitorOSD._rules_set:
            return
        try:
            from .hyprland import HyprlandIPC
            t = OSD_WINDOW_TITLE
            HyprlandIPC().batch([
                f"keyword windowrulev2 noanim, title:^{t}$",
                f"keyword windowrulev2 nofocus, title:^{t}$",
                f"keyword windowrulev2 noshadow, title:^{t}$",
                f"keyword windowrulev2 noborder, title:^{t}$",
            ])
            MonitorOSD._rules_set = True
        except Exception:
            pass

    def dismiss(self) -> None:
        if self._timer_id:
            GLib.source_remove(self._timer_id)
            self._timer_id = 0
        self.close()

    def _dismiss(self) -> bool:
        self._timer_id = 0
        self.close()
        return False


class MainWindow(Adw.ApplicationWindow):
    """Main window with canvas, properties panel, and toolbar."""

    __gtype_name__ = "MainWindow"

    def __init__(self, app: Adw.Application) -> None:
        super().__init__(application=app, title="Monique", default_width=1100, default_height=700)
        self._ipc = HyprlandIPC()
        self._profile_mgr = ProfileManager()
        self._monitors: list[MonitorConfig] = []
        self._workspace_rules = []
        self._current_profile_name: str = ""
        self._confirm_timer_id: int = 0
        self._inhibit_profile_switch: bool = False
        self._osd: MonitorOSD | None = None
        self._dirty: bool = False
        self._inhibit_clamshell_toggle: bool = False
        self._app_settings = load_app_settings()

        self._build_ui()
        self._setup_actions()
        self._load_current_state()
        self.connect("close-request", self._on_close_request)

    # ── UI Construction ──────────────────────────────────────────────

    def _build_ui(self) -> None:
        # Main vertical box
        main_box = Gtk.Box(orientation=Gtk.Orientation.VERTICAL)
        self.set_content(main_box)

        # Header bar
        header = Adw.HeaderBar()
        main_box.append(header)

        # Profile dropdown
        self._profile_dropdown = Gtk.DropDown()
        self._profile_dropdown.set_tooltip_text("Select profile")
        self._refresh_profile_list()
        self._profile_dropdown.connect("notify::selected", self._on_profile_selected)
        header.pack_start(self._profile_dropdown)

        # Save button
        btn_save = Gtk.Button(icon_name="document-save-symbolic", tooltip_text="Save profile")
        btn_save.connect("clicked", self._on_save_clicked)
        header.pack_start(btn_save)

        # Delete profile button
        btn_del = Gtk.Button(icon_name="edit-delete-symbolic", tooltip_text="Delete profile")
        btn_del.connect("clicked", self._on_delete_profile_clicked)
        header.pack_start(btn_del)

        # Spacer / title
        header.set_title_widget(Adw.WindowTitle(title="Monique", subtitle="Monitor Configuration"))

        # Apply button
        btn_apply = Gtk.Button(label="Apply", tooltip_text="Apply configuration")
        btn_apply.add_css_class("suggested-action")
        btn_apply.connect("clicked", self._on_apply_clicked)
        header.pack_end(btn_apply)

        # Workspace button
        btn_ws = Gtk.Button(icon_name="view-grid-symbolic", tooltip_text="Workspace rules")
        btn_ws.connect("clicked", self._on_workspaces_clicked)
        header.pack_end(btn_ws)

        # Detect monitors button
        btn_detect = Gtk.Button(icon_name="view-refresh-symbolic", tooltip_text="Detect monitors")
        btn_detect.connect("clicked", self._on_detect_clicked)
        header.pack_end(btn_detect)

        # Clamshell mode toggle button
        self._btn_clamshell = Gtk.ToggleButton(
            icon_name="computer-symbolic",
            tooltip_text="Clamshell mode (disable internal display)",
        )
        self._btn_clamshell.connect("toggled", self._on_clamshell_toggled)
        header.pack_end(self._btn_clamshell)

        # Hamburger menu
        self._build_menu(header)

        # Split view: canvas + properties
        self._split = Adw.OverlaySplitView()
        self._split.set_collapsed(False)
        self._split.set_sidebar_position(Gtk.PackType.END)
        self._split.set_max_sidebar_width(380)
        self._split.set_min_sidebar_width(300)
        main_box.append(self._split)

        # Canvas
        self._canvas = MonitorCanvas()
        self._canvas.set_use_description(not self._app_settings.get("use_port_names", False))
        self._canvas.connect("monitor-selected", self._on_monitor_selected)
        self._canvas.connect("monitor-moved", self._on_monitor_moved)
        self._canvas.connect("monitor-double-clicked", self._on_monitor_double_clicked)
        self._split.set_content(self._canvas)

        # Properties panel in a scrolled window
        scroll = Gtk.ScrolledWindow()
        scroll.set_policy(Gtk.PolicyType.NEVER, Gtk.PolicyType.AUTOMATIC)
        self._props = PropertiesPanel()
        self._props.connect("property-changed", self._on_property_changed)
        scroll.set_child(self._props)
        self._split.set_sidebar(scroll)

        # Status bar
        self._status = Gtk.Label(label="Ready", xalign=0)
        self._status.set_margin_start(12)
        self._status.set_margin_end(12)
        self._status.set_margin_top(4)
        self._status.set_margin_bottom(4)
        self._status.add_css_class("dim-label")
        main_box.append(self._status)

        # Toast overlay (wraps the split view)
        self._toast_overlay = Adw.ToastOverlay()
        main_box.remove(self._split)
        self._toast_overlay.set_child(self._split)
        # Re-insert toast overlay where split was
        main_box.remove(self._status)
        main_box.append(self._toast_overlay)
        main_box.append(self._status)

    def _setup_actions(self) -> None:
        """Set up keyboard shortcuts."""
        # Ctrl+S -> save
        action_save = Gio.SimpleAction(name="save")
        action_save.connect("activate", lambda *_: self._on_save_clicked(None))
        self.add_action(action_save)

        # Ctrl+R -> detect
        action_detect = Gio.SimpleAction(name="detect")
        action_detect.connect("activate", lambda *_: self._on_detect_clicked(None))
        self.add_action(action_detect)

        app = self.get_application()
        app.set_accels_for_action("win.save", ["<Control>s"])
        app.set_accels_for_action("win.detect", ["<Control>r"])

    def _build_menu(self, header: Adw.HeaderBar) -> None:
        """Build the hamburger menu with sections."""
        menu = Gio.Menu()

        # ── Settings section ──
        section_settings = Gio.Menu()
        item_prefs = Gio.MenuItem.new("Preferences", "win.preferences")
        item_prefs.set_icon(Gio.ThemedIcon.new("preferences-system-symbolic"))
        section_settings.append_item(item_prefs)
        menu.append_section(None, section_settings)

        if is_sddm_running():
            # SDDM toggle action (used in preferences dialog)
            update_sddm = self._app_settings.get("update_sddm", True)
            action_sddm = Gio.SimpleAction.new_stateful(
                "update-sddm", None, GLib.Variant.new_boolean(update_sddm),
            )
            action_sddm.connect("activate", self._on_sddm_toggled)
            self.add_action(action_sddm)

        # Preferences action
        action_prefs = Gio.SimpleAction(name="preferences")
        action_prefs.connect("activate", self._on_preferences)
        self.add_action(action_prefs)

        # ── About section ──
        section_about = Gio.Menu()
        item_about = Gio.MenuItem.new("About Monique", "win.about")
        item_about.set_icon(Gio.ThemedIcon.new("help-about-symbolic"))
        section_about.append_item(item_about)
        menu.append_section(None, section_about)

        action_about = Gio.SimpleAction(name="about")
        action_about.connect("activate", self._on_about)
        self.add_action(action_about)

        menu_btn = Gtk.MenuButton(
            icon_name="open-menu-symbolic",
            menu_model=menu,
            tooltip_text="Menu",
        )
        header.pack_end(menu_btn)

    def _on_preferences(self, action: Gio.SimpleAction, param) -> None:
        """Show the preferences dialog."""
        dialog = Adw.PreferencesDialog(title="Preferences")

        # ── Display Manager group ──
        page = Adw.PreferencesPage(
            title="General",
            icon_name="preferences-system-symbolic",
        )
        dialog.add(page)

        # ── Monitor Identification group ──
        grp_id = Adw.PreferencesGroup(
            title="Monitor Identification",
            description="How monitors are identified in config files",
        )
        page.add(grp_id)

        sw_port = Adw.SwitchRow(
            title="Use port names (e.g. DP-1, HDMI-A-1)",
            subtitle="When disabled, monitors are identified by description (recommended)",
        )
        sw_port.set_active(self._app_settings.get("use_port_names", False))
        sw_port.connect("notify::active", self._on_port_names_switch_changed)
        grp_id.add(sw_port)

        # ── Display Manager group ──
        grp_dm = Adw.PreferencesGroup(
            title="Display Manager",
            description="Configure login screen integration",
        )
        page.add(grp_dm)

        has_dm = False

        if is_sddm_running():
            has_dm = True
            sw_sddm = Adw.SwitchRow(
                title="Update SDDM Xsetup",
                subtitle="Write xrandr layout to the SDDM login screen script",
                icon_name="system-lock-screen-symbolic",
            )
            sw_sddm.set_active(self._app_settings.get("update_sddm", True))
            sw_sddm.connect("notify::active", self._on_sddm_switch_changed)
            grp_dm.add(sw_sddm)

        if is_greetd_running():
            has_dm = True
            sw_greetd = Adw.SwitchRow(
                title="Update greetd sway config",
                subtitle="Add 'include /etc/greetd/monique-monitors.conf' to your sway-config",
                icon_name="system-lock-screen-symbolic",
            )
            sw_greetd.set_active(self._app_settings.get("update_greetd", True))
            sw_greetd.connect("notify::active", self._on_greetd_switch_changed)
            grp_dm.add(sw_greetd)

        if not has_dm:
            row_no_dm = Adw.ActionRow(
                title="No supported display manager detected",
                icon_name="dialog-information-symbolic",
            )
            row_no_dm.add_css_class("dim-label")
            grp_dm.add(row_no_dm)

        # ── Workspaces group ──
        grp_ws = Adw.PreferencesGroup(
            title="Workspaces",
            description="Configure workspace behavior on monitor changes",
        )
        page.add(grp_ws)

        sw_migrate = Adw.SwitchRow(
            title="Migrate workspaces on monitor removal",
            subtitle="Move workspaces to primary monitor when their monitor is disabled",
        )
        sw_migrate.set_active(self._app_settings.get("migrate_workspaces", True))
        sw_migrate.connect("notify::active", self._on_migrate_switch_changed)
        grp_ws.add(sw_migrate)

        # ── Clamshell Mode group ──
        grp_clam = Adw.PreferencesGroup(
            title="Clamshell Mode",
            description="Automatically disable the internal display when external monitors are connected",
        )
        page.add(grp_clam)

        sw_clamshell = Adw.SwitchRow(
            title="Enable clamshell mode",
            subtitle="Disable laptop screen when external monitors are present (daemon)",
            icon_name="computer-symbolic",
        )
        sw_clamshell.set_active(self._app_settings.get("clamshell_mode", False))
        sw_clamshell.connect("notify::active", self._on_clamshell_switch_changed)
        grp_clam.add(sw_clamshell)

        dialog.present(self)

    def _on_sddm_switch_changed(self, row: Adw.SwitchRow, pspec) -> None:
        """Handle the SDDM switch toggle in preferences."""
        new_val = row.get_active()
        self._app_settings["update_sddm"] = new_val
        save_app_settings(self._app_settings)
        # Sync the stateful action if it exists
        action = self.lookup_action("update-sddm")
        if action:
            action.set_state(GLib.Variant.new_boolean(new_val))
        state = "enabled" if new_val else "disabled"
        self._toast(f"SDDM Xsetup update {state}")

    def _on_greetd_switch_changed(self, row: Adw.SwitchRow, pspec) -> None:
        """Handle the greetd switch toggle in preferences."""
        new_val = row.get_active()
        self._app_settings["update_greetd"] = new_val
        save_app_settings(self._app_settings)
        state = "enabled" if new_val else "disabled"
        self._toast(f"greetd monitor config update {state}")

    def _on_migrate_switch_changed(self, row: Adw.SwitchRow, pspec) -> None:
        """Handle the migrate workspaces switch toggle in preferences."""
        new_val = row.get_active()
        self._app_settings["migrate_workspaces"] = new_val
        save_app_settings(self._app_settings)
        state = "enabled" if new_val else "disabled"
        self._toast(f"Workspace migration {state}")

    def _on_port_names_switch_changed(self, row: Adw.SwitchRow, pspec) -> None:
        """Handle the port names switch toggle in preferences."""
        new_val = row.get_active()
        self._app_settings["use_port_names"] = new_val
        save_app_settings(self._app_settings)
        self._canvas.set_use_description(not new_val)
        mode = "port names" if new_val else "descriptions"
        self._toast(f"Monitor identification: {mode}")

    def _on_clamshell_switch_changed(self, row: Adw.SwitchRow, pspec) -> None:
        """Handle the clamshell mode switch toggle in preferences."""
        new_val = row.get_active()
        self._app_settings["clamshell_mode"] = new_val
        save_app_settings(self._app_settings)
        state = "enabled" if new_val else "disabled"
        self._toast(f"Clamshell mode {state}")

    def _on_clamshell_toggled(self, btn: Gtk.ToggleButton) -> None:
        """Handle the clamshell toggle button in the headerbar."""
        if self._inhibit_clamshell_toggle:
            return

        internals = [m for m in self._monitors if m.is_internal]
        externals = [m for m in self._monitors if not m.is_internal]

        if btn.get_active():
            # Activate clamshell
            if not internals:
                self._toast("No internal display detected")
                self._inhibit_clamshell_toggle = True
                btn.set_active(False)
                self._inhibit_clamshell_toggle = False
                return
            if not any(m.enabled for m in externals):
                self._toast("No external monitor enabled")
                self._inhibit_clamshell_toggle = True
                btn.set_active(False)
                self._inhibit_clamshell_toggle = False
                return
            if apply_clamshell(self._monitors):
                self._canvas.monitors = self._monitors
                self._place_disabled_monitors()
                self._canvas.queue_draw()
                self._update_properties_for_selected()
                self._on_apply_clicked(None)
        else:
            # Deactivate clamshell
            if undo_clamshell(self._monitors):
                self._canvas.monitors = self._monitors
                self._place_disabled_monitors()
                self._canvas.queue_draw()
                self._update_properties_for_selected()
                self._on_apply_clicked(None)

    def _on_sddm_toggled(self, action: Gio.SimpleAction, param) -> None:
        """Toggle the 'update SDDM Xsetup' setting (from menu action)."""
        new_val = not action.get_state().get_boolean()
        action.set_state(GLib.Variant.new_boolean(new_val))
        self._app_settings["update_sddm"] = new_val
        save_app_settings(self._app_settings)
        state = "enabled" if new_val else "disabled"
        self._toast(f"SDDM Xsetup update {state}")

    def _on_about(self, action: Gio.SimpleAction, param) -> None:
        """Show the About dialog."""
        about = Adw.AboutDialog(
            application_name="Monique",
            application_icon="com.github.monique",
            version="0.3.0",
            developer_name="Monique contributors",
            comments="MONitor Integrated QUick Editor for Hyprland and Sway",
            license_type=Gtk.License.GPL_3_0,
        )
        about.present(self)

    # ── Data Loading ─────────────────────────────────────────────────

    def _load_workspace_rules_from_conf(self) -> list[WorkspaceRule]:
        """Read workspace rules from the monitors.conf file we wrote."""
        conf = hyprland_config_dir() / "monitors.conf"
        if not conf.exists():
            return []
        rules: list[WorkspaceRule] = []
        for line in conf.read_text(encoding="utf-8").splitlines():
            rule = WorkspaceRule.from_hyprland_line(line)
            if rule:
                rules.append(rule)
        return rules

    def _load_current_state(self) -> None:
        """Query Hyprland for current monitor state."""
        try:
            self._monitors = self._ipc.get_monitors()
            self._place_disabled_monitors()
            self._workspace_rules = self._load_workspace_rules_from_conf()
            self._canvas.monitors = self._monitors
            if self._monitors:
                self._canvas.selected_index = 0
                self._update_properties_for_selected()
            n_enabled = sum(1 for m in self._monitors if m.enabled)
            n_disabled = sum(1 for m in self._monitors if not m.enabled)
            parts = [f"{n_enabled} active"]
            if n_disabled:
                parts.append(f"{n_disabled} disabled")
            self._set_status(f"{len(self._monitors)} monitor(s) detected ({', '.join(parts)})")
            # Sync clamshell toggle: active if an internal is disabled while externals are present
            has_disabled_internal = any(m.is_internal and not m.enabled for m in self._monitors)
            has_enabled_external = any(not m.is_internal and m.enabled for m in self._monitors)
            self._inhibit_clamshell_toggle = True
            self._btn_clamshell.set_active(has_disabled_internal and has_enabled_external)
            self._inhibit_clamshell_toggle = False
            self._select_matching_profile()
        except Exception as e:
            self._set_status(f"Error: {e}")
            self._toast(f"Cannot connect to Hyprland: {e}")

    def _place_disabled_monitors(self) -> None:
        """Position disabled monitors below the active layout so they're visible."""
        enabled = [m for m in self._monitors if m.enabled]
        disabled = [m for m in self._monitors if not m.enabled]
        if not disabled:
            return

        if enabled:
            max_y = max(m.y + m.logical_height for m in enabled)
            min_x = min(m.x for m in enabled)
        else:
            max_y = 0
            min_x = 0

        gap = 200  # logical pixels gap between active and disabled section
        x_cursor = min_x
        for m in disabled:
            m.x = round(x_cursor)
            m.y = round(max_y + gap)
            x_cursor += m.logical_width + 100

    def _update_properties_for_selected(self) -> None:
        idx = self._canvas.selected_index
        if 0 <= idx < len(self._monitors):
            m = self._monitors[idx]
            names = [mon.name for mon in self._monitors]
            self._props.set_mirror_monitors(names)
            self._props.update_from_monitor(m)
        else:
            self._props.update_from_monitor(None)

    def _select_matching_profile(self) -> None:
        """If the current monitor layout matches a saved profile, select it."""
        current_fp = sorted(m.description for m in self._monitors if m.description)
        match = self._profile_mgr.find_best_match(current_fp)
        if match is None:
            return
        model = self._profile_dropdown.get_model()
        for i in range(model.get_n_items()):
            if model.get_string(i) == match.name:
                self._inhibit_profile_switch = True
                self._profile_dropdown.set_selected(i)
                self._current_profile_name = match.name
                self._workspace_rules = match.workspace_rules
                self._inhibit_profile_switch = False
                break

    # ── Profile Management ───────────────────────────────────────────

    def _refresh_profile_list(self) -> None:
        names = self._profile_mgr.list_profiles()
        options = ["(Current)"] + names
        self._profile_dropdown.set_model(Gtk.StringList.new(options))
        self._profile_dropdown.set_selected(0)

    def _on_profile_selected(self, dropdown: Gtk.DropDown, pspec) -> None:
        if self._inhibit_profile_switch:
            return
        sel = dropdown.get_selected()
        if sel == 0:
            # Current = reload from Hyprland
            self._load_current_state()
            return
        model = dropdown.get_model()
        name = model.get_string(sel)
        profile = self._profile_mgr.load(name)
        if profile:
            self._monitors = profile.monitors
            self._workspace_rules = profile.workspace_rules
            self._current_profile_name = profile.name
            self._canvas.monitors = self._monitors
            if self._monitors:
                self._canvas.selected_index = 0
                self._update_properties_for_selected()
            self._set_status(f"Loaded profile: {name}")

    def _on_save_clicked(self, btn) -> None:
        dialog = Adw.AlertDialog()
        dialog.set_heading("Save Profile")
        dialog.add_response("cancel", "Cancel")
        dialog.add_response("save", "Save")
        dialog.set_response_appearance("save", Adw.ResponseAppearance.SUGGESTED)
        dialog.set_default_response("save")

        # Build content: radio for existing profiles + new name entry
        box = Gtk.Box(orientation=Gtk.Orientation.VERTICAL, spacing=8)
        box.set_margin_start(24)
        box.set_margin_end(24)

        existing = self._profile_mgr.list_profiles()
        self._save_radios: list[Gtk.CheckButton] = []
        group_leader: Gtk.CheckButton | None = None

        # Existing profiles as radio buttons
        if existing:
            label_existing = Gtk.Label(label="Overwrite existing:", xalign=0)
            label_existing.add_css_class("dim-label")
            box.append(label_existing)

            for name in existing:
                radio = Gtk.CheckButton(label=name)
                if group_leader is None:
                    group_leader = radio
                else:
                    radio.set_group(group_leader)
                self._save_radios.append(radio)
                box.append(radio)

            sep = Gtk.Separator()
            sep.set_margin_top(4)
            sep.set_margin_bottom(4)
            box.append(sep)

        # New profile option
        new_box = Gtk.Box(orientation=Gtk.Orientation.HORIZONTAL, spacing=8)
        self._save_radio_new = Gtk.CheckButton(label="New:")
        if group_leader is not None:
            self._save_radio_new.set_group(group_leader)
        else:
            group_leader = self._save_radio_new
        new_box.append(self._save_radio_new)

        self._save_entry = Gtk.Entry(hexpand=True)
        self._save_entry.set_placeholder_text("Profile name")
        self._save_entry.set_text(self._generate_profile_name())
        new_box.append(self._save_entry)
        box.append(new_box)

        # Auto-select: if editing a loaded profile, pre-select it; else select "New"
        preselected = False
        if self._current_profile_name and existing:
            for radio in self._save_radios:
                if radio.get_label() == self._current_profile_name:
                    radio.set_active(True)
                    preselected = True
                    break
        if not preselected:
            self._save_radio_new.set_active(True)

        # Focus entry when "New" is selected
        self._save_radio_new.connect("toggled", self._on_save_new_toggled)
        self._save_entry.connect("changed", lambda e: self._save_radio_new.set_active(True))

        dialog.set_extra_child(box)
        dialog.connect("response", self._on_save_response)
        dialog.present(self)

    def _on_save_new_toggled(self, radio: Gtk.CheckButton) -> None:
        if radio.get_active():
            self._save_entry.grab_focus()

    def _on_save_response(self, dialog: Adw.AlertDialog, response: str) -> None:
        if response != "save":
            return

        # Determine selected name
        name = ""
        if self._save_radio_new.get_active():
            name = self._save_entry.get_text().strip()
        else:
            for radio in self._save_radios:
                if radio.get_active():
                    name = radio.get_label()
                    break

        if not name:
            self._toast("No profile name specified")
            return

        profile = Profile(
            name=name,
            monitors=list(self._monitors),
            workspace_rules=list(self._workspace_rules),
        )
        self._profile_mgr.save(profile)
        self._current_profile_name = name
        self._inhibit_profile_switch = True
        self._refresh_profile_list()
        # Select the saved profile in the dropdown
        model = self._profile_dropdown.get_model()
        for i in range(model.get_n_items()):
            if model.get_string(i) == name:
                self._profile_dropdown.set_selected(i)
                break
        self._inhibit_profile_switch = False
        self._dirty = False
        self._toast(f"Profile '{name}' saved")

    def _on_delete_profile_clicked(self, btn) -> None:
        sel = self._profile_dropdown.get_selected()
        if sel == 0:
            self._toast("Cannot delete (Current)")
            return
        model = self._profile_dropdown.get_model()
        name = model.get_string(sel)
        self._profile_mgr.delete(name)
        self._refresh_profile_list()
        self._toast(f"Profile '{name}' deleted")

    def _generate_profile_name(self) -> str:
        """Generate a default profile name from connected monitors."""
        if not self._monitors:
            return "Default"
        names = [m.model or m.name for m in self._monitors]
        return " + ".join(names)

    # ── Dirty tracking ─────────────────────────────────────────────

    def _mark_dirty(self) -> None:
        """Switch dropdown to (Current) when the user modifies anything."""
        self._dirty = True
        if self._inhibit_profile_switch:
            return
        sel = self._profile_dropdown.get_selected()
        if sel != 0:
            self._inhibit_profile_switch = True
            self._profile_dropdown.set_selected(0)
            self._inhibit_profile_switch = False
        self._current_profile_name = ""

    # ── Canvas Events ────────────────────────────────────────────────

    def _on_monitor_selected(self, canvas: MonitorCanvas, index: int) -> None:
        self._update_properties_for_selected()

    def _on_monitor_double_clicked(self, canvas: MonitorCanvas, index: int) -> None:
        self._show_osd(index)

    def _on_monitor_moved(self, canvas: MonitorCanvas, index: int) -> None:
        self._mark_dirty()
        self._update_properties_for_selected()

    def _on_property_changed(self, panel: PropertiesPanel) -> None:
        self._mark_dirty()
        self._canvas.queue_draw()

    # ── Workspace Dialog ─────────────────────────────────────────────

    def _on_workspaces_clicked(self, btn) -> None:
        names = [m.name for m in self._monitors]
        descs = [m.description for m in self._monitors]
        enabled = [m.enabled for m in self._monitors]
        dialog = WorkspacePanel(
            monitor_names=names,
            monitor_descriptions=descs,
            monitor_enabled=enabled,
            application=self.get_application(),
        )
        dialog.set_rules(self._workspace_rules)
        dialog.connect("rules-changed", self._on_workspace_rules_changed)
        dialog.set_transient_for(self)
        dialog.present()

    def _on_workspace_rules_changed(self, panel: WorkspacePanel) -> None:
        self._workspace_rules = panel.get_rules()
        self._mark_dirty()

    # ── Detect ───────────────────────────────────────────────────────

    def _on_detect_clicked(self, btn) -> None:
        self._load_current_state()

    # ── Apply ────────────────────────────────────────────────────────

    def _on_apply_clicked(self, btn) -> None:
        """Apply the current configuration with backup and confirmation."""
        profile = Profile(
            name=self._current_profile_name or "applied",
            monitors=list(self._monitors),
            workspace_rules=list(self._workspace_rules),
        )

        conf_dir = hyprland_config_dir()
        monitors_conf = conf_dir / "monitors.conf"

        # Snapshot workspaces before applying
        self._ws_snapshot = self._ipc.get_workspaces()
        self._migrated_workspaces: list[tuple[str, str]] = []

        # Backup
        backup_file(monitors_conf)

        try:
            update_sddm = self._app_settings.get("update_sddm", True)
            update_greetd = self._app_settings.get("update_greetd", True)
            use_desc = not self._app_settings.get("use_port_names", False)
            self._ipc.apply_profile(
                profile, update_sddm=update_sddm, update_greetd=update_greetd,
                use_description=use_desc,
            )
            self._set_status("Configuration applied")
        except Exception as e:
            self._toast(f"Apply failed: {e}")
            restore_backup(monitors_conf)
            return

        # Migrate orphaned workspaces if setting is on
        if self._app_settings.get("migrate_workspaces", True):
            self._migrate_orphaned_workspaces(profile)

        # Show confirmation dialog with countdown
        self._show_confirm_dialog(monitors_conf)

    def _migrate_orphaned_workspaces(self, profile: Profile) -> None:
        """Move workspaces from disabled/removed monitors to the primary monitor."""
        enabled_names = {m.name for m in profile.monitors if m.enabled}
        if not enabled_names:
            return

        # Primary = first enabled monitor in profile
        primary = next(m.name for m in profile.monitors if m.enabled)

        self._migrated_workspaces = []
        for ws in self._ws_snapshot:
            ws_monitor = ws.get("monitor", "")
            ws_name = str(ws.get("name", ws.get("id", "")))
            if ws_monitor and ws_monitor not in enabled_names:
                try:
                    self._ipc.move_workspace_to_monitor(ws_name, primary)
                    self._migrated_workspaces.append((ws_name, ws_monitor))
                except Exception:
                    pass

    def _show_confirm_dialog(self, conf_path) -> None:
        self._confirm_remaining = CONFIRM_TIMEOUT
        dialog = Adw.AlertDialog()
        dialog.set_heading("Keep Settings?")
        dialog.set_body(f"Reverting in {self._confirm_remaining}s if not confirmed...")
        dialog.add_response("revert", "Revert")
        dialog.set_response_appearance("revert", Adw.ResponseAppearance.DESTRUCTIVE)
        dialog.add_response("keep", "Keep Settings")
        dialog.set_response_appearance("keep", Adw.ResponseAppearance.SUGGESTED)
        dialog.set_default_response("keep")
        dialog.set_close_response("revert")

        self._confirm_dialog = dialog
        self._confirm_conf_path = conf_path

        # Start countdown
        self._confirm_timer_id = GLib.timeout_add(1000, self._confirm_tick)
        dialog.connect("response", self._on_confirm_response)
        dialog.present(self)

    def _confirm_tick(self) -> bool:
        self._confirm_remaining -= 1
        if self._confirm_remaining <= 0:
            # Auto-revert
            self._confirm_dialog.force_close()
            self._do_revert()
            return False
        self._confirm_dialog.set_body(f"Reverting in {self._confirm_remaining}s if not confirmed...")
        return True

    def _on_confirm_response(self, dialog: Adw.AlertDialog, response: str) -> None:
        if self._confirm_timer_id:
            GLib.source_remove(self._confirm_timer_id)
            self._confirm_timer_id = 0

        if response == "keep":
            # Remove backup
            bak = self._confirm_conf_path.with_suffix(self._confirm_conf_path.suffix + ".bak")
            if bak.exists():
                bak.unlink()
            self._migrated_workspaces = []
            self._toast("Settings kept")
        else:
            self._do_revert()

    def _do_revert(self) -> None:
        # Restore migrated workspaces to their original monitors
        for ws_name, original_monitor in self._migrated_workspaces:
            try:
                self._ipc.move_workspace_to_monitor(ws_name, original_monitor)
            except Exception:
                pass
        self._migrated_workspaces = []

        if restore_backup(self._confirm_conf_path):
            try:
                self._ipc.reload()
                self._load_current_state()
                self._toast("Settings reverted")
            except Exception as e:
                self._toast(f"Revert reload failed: {e}")
        else:
            self._toast("No backup to revert")

    # ── OSD ───────────────────────────────────────────────────────────

    def _show_osd(self, index: int) -> None:
        """Show an OSD overlay on the physical monitor corresponding to index."""
        # Dismiss previous
        if self._osd is not None:
            self._osd.dismiss()
            self._osd = None

        if index < 0 or index >= len(self._monitors):
            return

        mon = self._monitors[index]
        gdk_mon = self._find_gdk_monitor(mon.name)
        if gdk_mon is None:
            return

        self._osd = MonitorOSD(self.get_application(), gdk_mon, mon.name)
        self._osd.show_timed()

        # Re-focus the main window so the OSD doesn't steal it
        GLib.idle_add(self.present)

    def _find_gdk_monitor(self, connector: str) -> Gdk.Monitor | None:
        """Find the Gdk.Monitor matching a connector name like 'DP-1'."""
        display = Gdk.Display.get_default()
        if display is None:
            return None
        monitors = display.get_monitors()
        for i in range(monitors.get_n_items()):
            gdk_mon = monitors.get_item(i)
            if gdk_mon.get_connector() == connector:
                return gdk_mon
        return None

    # ── Close ─────────────────────────────────────────────────────────

    def _on_close_request(self, window: Adw.ApplicationWindow) -> bool:
        if not self._dirty:
            return False  # allow close

        dialog = Adw.AlertDialog()
        dialog.set_heading("Unsaved Changes")
        dialog.set_body("Save changes before closing?")
        dialog.add_response("discard", "Discard")
        dialog.set_response_appearance("discard", Adw.ResponseAppearance.DESTRUCTIVE)
        dialog.add_response("cancel", "Cancel")
        dialog.add_response("save", "Save")
        dialog.set_response_appearance("save", Adw.ResponseAppearance.SUGGESTED)
        dialog.set_default_response("save")
        dialog.set_close_response("cancel")
        dialog.connect("response", self._on_close_dialog_response)
        dialog.present(self)
        return True  # block close for now

    def _on_close_dialog_response(self, dialog: Adw.AlertDialog, response: str) -> None:
        if response == "cancel":
            return
        if response == "save":
            self._on_save_clicked(None)
            # Close after save dialog completes (dirty will be False)
            return
        # discard
        self._dirty = False
        self.close()

    # ── Helpers ──────────────────────────────────────────────────────

    def _set_status(self, text: str) -> None:
        self._status.set_text(text)

    def _toast(self, message: str) -> None:
        toast = Adw.Toast(title=message, timeout=3)
        self._toast_overlay.add_toast(toast)



"""Data models: MonitorConfig, WorkspaceRule, Profile."""

from __future__ import annotations

from dataclasses import dataclass, field, asdict
from enum import Enum
from typing import ClassVar


# ── Enums ────────────────────────────────────────────────────────────────

class ResolutionMode(Enum):
    EXPLICIT = "explicit"
    PREFERRED = "preferred"
    HIGHRES = "highres"
    HIGHRR = "highrr"


class PositionMode(Enum):
    EXPLICIT = "explicit"
    AUTO = "auto"
    AUTO_RIGHT = "auto-right"
    AUTO_LEFT = "auto-left"
    AUTO_UP = "auto-up"
    AUTO_DOWN = "auto-down"
    AUTO_CENTER_RIGHT = "auto-center-right"
    AUTO_CENTER_LEFT = "auto-center-left"
    AUTO_CENTER_UP = "auto-center-up"
    AUTO_CENTER_DOWN = "auto-center-down"


class ScaleMode(Enum):
    EXPLICIT = "explicit"
    AUTO = "auto"


class VRR(Enum):
    OFF = 0
    ON = 1
    FULLSCREEN = 2


class Transform(Enum):
    NORMAL = 0
    ROTATE_90 = 1
    ROTATE_180 = 2
    ROTATE_270 = 3
    FLIPPED = 4
    FLIPPED_90 = 5
    FLIPPED_180 = 6
    FLIPPED_270 = 7

    @property
    def label(self) -> str:
        labels = {
            0: "Normal",
            1: "90°",
            2: "180°",
            3: "270°",
            4: "Flipped",
            5: "Flipped 90°",
            6: "Flipped 180°",
            7: "Flipped 270°",
        }
        return labels[self.value]

    @property
    def is_rotated(self) -> bool:
        """True if width/height are swapped (90° or 270° variants)."""
        return self.value in (1, 3, 5, 7)


# ── MonitorConfig ────────────────────────────────────────────────────────

@dataclass
class MonitorConfig:
    # Identity (from hyprctl monitors -j)
    name: str = ""              # e.g. "DP-1", "HDMI-A-1"
    description: str = ""       # e.g. "LG Electronics LG ULTRAWIDE 0x00038C43"
    make: str = ""
    model: str = ""
    serial: str = ""

    # Resolution
    width: int = 1920
    height: int = 1080
    refresh_rate: float = 60.0
    resolution_mode: ResolutionMode = ResolutionMode.EXPLICIT

    # Available modes from hardware (list of "WxH@R" strings)
    available_modes: list[str] = field(default_factory=list)

    # Position
    x: int = 0
    y: int = 0
    position_mode: PositionMode = PositionMode.AUTO

    # Scale
    scale: float = 1.0
    scale_mode: ScaleMode = ScaleMode.EXPLICIT

    # Transform
    transform: Transform = Transform.NORMAL

    # Mirror
    mirror_of: str = ""

    # Advanced
    bitdepth: int = 8           # 8 or 10
    vrr: VRR = VRR.OFF
    color_management: str = ""  # "", "srgb", "dcip3", "dp3", "adobe", "wide", "edid", "hdr", "hdredid"
    sdr_brightness: float = 1.0
    sdr_saturation: float = 1.0

    # Reserved area
    reserved_top: int = 0
    reserved_bottom: int = 0
    reserved_left: int = 0
    reserved_right: int = 0

    # Enabled
    enabled: bool = True

    @property
    def logical_width(self) -> float:
        """Width in logical pixels (accounting for scale and rotation)."""
        w, h = self.width, self.height
        if self.transform.is_rotated:
            w, h = h, w
        return w / self.scale

    @property
    def logical_height(self) -> float:
        """Height in logical pixels (accounting for scale and rotation)."""
        w, h = self.width, self.height
        if self.transform.is_rotated:
            w, h = h, w
        return h / self.scale

    @property
    def is_internal(self) -> bool:
        """True if this is a built-in laptop display (eDP or LVDS port)."""
        prefix = self.name.split("-")[0].upper() if self.name else ""
        return prefix in ("EDP", "LVDS", "DSI")

    # Mapping from Transform enum to xrandr --rotate / --reflect values
    # Each entry is (rotate, reflect_or_None)
    _XRANDR_TRANSFORMS: ClassVar[dict[int, tuple[str, str | None]]] = {
        0: ("normal", None),
        1: ("left", None),
        2: ("inverted", None),
        3: ("right", None),
        4: ("normal", "x"),
        5: ("left", "x"),
        6: ("inverted", "x"),
        7: ("right", "x"),
    }

    # Mapping from Transform enum to Sway transform strings
    _SWAY_TRANSFORMS: ClassVar[dict[int, str]] = {
        0: "normal",
        1: "90",
        2: "180",
        3: "270",
        4: "flipped",
        5: "flipped-90",
        6: "flipped-180",
        7: "flipped-270",
    }

    # Inverse mapping: Sway transform string -> Transform enum value
    _SWAY_TRANSFORMS_INV: ClassVar[dict[str, int]] = {
        v: k for k, v in _SWAY_TRANSFORMS.items()
    }

    def to_sway_block(self, use_description: bool = False) -> str:
        """Generate the `output` config block for sway."""
        identifier = (
            f'"{self.description}"'
            if use_description and self.description
            else self.name
        )
        if not self.enabled:
            return f"output {identifier} disable"

        lines: list[str] = []

        # Resolution: only emit mode for explicit resolutions
        if self.resolution_mode == ResolutionMode.EXPLICIT:
            lines.append(f"    mode {self.width}x{self.height}@{self.refresh_rate:.3f}Hz")

        # Position: always emit explicit coordinates
        if self.position_mode == PositionMode.EXPLICIT:
            lines.append(f"    pos {self.x} {self.y}")

        # Scale
        if self.scale_mode == ScaleMode.EXPLICIT:
            lines.append(f"    scale {self.scale:g}")

        # Transform
        lines.append(f"    transform {self._SWAY_TRANSFORMS[self.transform.value]}")

        # VRR → adaptive_sync
        lines.append(f"    adaptive_sync {'on' if self.vrr != VRR.OFF else 'off'}")

        body = "\n".join(lines)
        return f"output {identifier} {{\n{body}\n}}"

    def to_xrandr_cmd(self) -> str:
        """Generate an xrandr command line for this monitor."""
        if not self.enabled:
            return f"xrandr --output {self.name} --off"

        parts = [f"xrandr --output {self.name}"]

        # Resolution
        if self.resolution_mode == ResolutionMode.EXPLICIT:
            parts.append(f"--mode {self.width}x{self.height}")
            parts.append(f"--rate {self.refresh_rate:.2f}")
        else:
            parts.append("--auto")

        # Position
        if self.position_mode == PositionMode.EXPLICIT:
            parts.append(f"--pos {self.x}x{self.y}")

        # Transform (rotate + reflect)
        rotate, reflect = self._XRANDR_TRANSFORMS[self.transform.value]
        parts.append(f"--rotate {rotate}")
        if reflect:
            parts.append(f"--reflect {reflect}")

        return " ".join(parts)

    def to_hyprland_line(
        self,
        use_description: bool = False,
        name_to_id: dict[str, str] | None = None,
    ) -> str:
        """Generate the `monitor=...` config line for hyprland.conf."""
        parts: list[str] = []

        # Name — use desc:DESCRIPTION when use_description is enabled
        if use_description and self.description:
            parts.append(f"desc:{self.description}")
        else:
            parts.append(self.name)

        # Disabled monitor
        if not self.enabled:
            parts.append("disable")
            return "monitor=" + ", ".join(parts)

        # Resolution
        if self.resolution_mode == ResolutionMode.EXPLICIT:
            refresh = f"{self.refresh_rate:g}"
            parts.append(f"{self.width}x{self.height}@{refresh}")
        else:
            parts.append(self.resolution_mode.value)

        # Position
        if self.position_mode == PositionMode.EXPLICIT:
            parts.append(f"{self.x}x{self.y}")
        else:
            parts.append(self.position_mode.value)

        # Scale
        if self.scale_mode == ScaleMode.AUTO:
            parts.append("auto")
        else:
            parts.append(f"{self.scale:g}")

        # Optional extras
        extras: list[str] = []

        if self.transform != Transform.NORMAL:
            extras.append(f"transform, {self.transform.value}")

        if self.mirror_of:
            mirror_id = self.mirror_of
            if name_to_id and self.mirror_of in name_to_id:
                mirror_id = name_to_id[self.mirror_of]
            extras.append(f"mirror, {mirror_id}")

        if self.bitdepth != 8:
            extras.append(f"bitdepth, {self.bitdepth}")

        if self.vrr != VRR.OFF:
            extras.append(f"vrr, {self.vrr.value}")

        if self.color_management:
            extras.append(f"cm, {self.color_management}")

        if self.sdr_brightness != 1.0:
            extras.append(f"sdrbrightness, {self.sdr_brightness:g}")

        if self.sdr_saturation != 1.0:
            extras.append(f"sdrsaturation, {self.sdr_saturation:g}")

        if any((self.reserved_top, self.reserved_bottom, self.reserved_left, self.reserved_right)):
            extras.append(
                f"addreserved, {self.reserved_top}, {self.reserved_bottom}, "
                f"{self.reserved_left}, {self.reserved_right}"
            )

        line = "monitor=" + ", ".join(parts)
        for extra in extras:
            line += f", {extra}"

        return line

    def to_dict(self) -> dict:
        """Serialize to a JSON-friendly dict."""
        d = asdict(self)
        d["resolution_mode"] = self.resolution_mode.value
        d["position_mode"] = self.position_mode.value
        d["scale_mode"] = self.scale_mode.value
        d["transform"] = self.transform.value
        d["vrr"] = self.vrr.value
        return d

    @classmethod
    def from_dict(cls, d: dict) -> MonitorConfig:
        """Deserialize from a dict."""
        d = dict(d)  # copy
        d["resolution_mode"] = ResolutionMode(d.get("resolution_mode", "explicit"))
        d["position_mode"] = PositionMode(d.get("position_mode", "auto"))
        d["scale_mode"] = ScaleMode(d.get("scale_mode", "explicit"))
        d["transform"] = Transform(d.get("transform", 0))
        d["vrr"] = VRR(d.get("vrr", 0))
        return cls(**{k: v for k, v in d.items() if k in cls.__dataclass_fields__})

    @classmethod
    def from_hyprctl(cls, data: dict) -> MonitorConfig:
        """Create from hyprctl monitors -j output."""
        modes = list(data.get("availableModes", []))

        # VRR: Hyprland JSON returns bool (false/true), map to 0/1
        vrr_raw = data.get("vrr", False)
        if isinstance(vrr_raw, bool):
            vrr_val = VRR.ON if vrr_raw else VRR.OFF
        elif isinstance(vrr_raw, int):
            vrr_val = VRR(vrr_raw)
        else:
            vrr_val = VRR.OFF

        # Note: reserved area from hyprctl reflects runtime state (bars etc.),
        # not user config. We don't import it to avoid writing addreserved
        # for values set by other programs.

        disabled = data.get("disabled", False)
        raw_x = data.get("x", 0)
        raw_y = data.get("y", 0)

        # Disabled monitors report x=-1, y=-1; use auto positioning
        if disabled and raw_x < 0 and raw_y < 0:
            pos_mode = PositionMode.AUTO
            raw_x = 0
            raw_y = 0
        else:
            pos_mode = PositionMode.EXPLICIT

        return cls(
            name=data.get("name", ""),
            description=data.get("description", ""),
            make=data.get("make", ""),
            model=data.get("model", ""),
            serial=data.get("serial", ""),
            width=data.get("width", 1920),
            height=data.get("height", 1080),
            refresh_rate=round(data.get("refreshRate", 60.0), 2),
            resolution_mode=ResolutionMode.EXPLICIT,
            available_modes=modes,
            x=raw_x,
            y=raw_y,
            position_mode=pos_mode,
            scale=data.get("scale", 1.0),
            scale_mode=ScaleMode.EXPLICIT,
            transform=Transform(data.get("transform", 0)),
            enabled=not disabled,
            vrr=vrr_val,
        )

    @classmethod
    def from_sway_output(cls, data: dict) -> MonitorConfig:
        """Create from swaymsg -t get_outputs JSON output."""
        # Sway doesn't have a 'description' field — reconstruct from components
        make = data.get("make", "")
        model = data.get("model", "")
        serial = data.get("serial", "")
        description = f"{make} {model} {serial}".strip()

        # Current mode
        current_mode = data.get("current_mode", {})
        width = current_mode.get("width", 1920)
        height = current_mode.get("height", 1080)
        # Sway reports refresh in millihertz
        refresh_mhz = current_mode.get("refresh", 60000)
        refresh_rate = round(refresh_mhz / 1000.0, 2)

        # Available modes
        modes: list[str] = []
        for m in data.get("modes", []):
            mw = m.get("width", 0)
            mh = m.get("height", 0)
            mr = round(m.get("refresh", 0) / 1000.0, 2)
            modes.append(f"{mw}x{mh}@{mr:.2f}Hz")

        # Position
        rect = data.get("rect", {})
        raw_x = rect.get("x", 0)
        raw_y = rect.get("y", 0)

        # Scale: -1 means output is disabled in Sway
        raw_scale = data.get("scale", 1.0)
        if raw_scale < 0:
            enabled = False
            scale = 1.0
            pos_mode = PositionMode.AUTO
            raw_x = 0
            raw_y = 0
        else:
            enabled = data.get("active", True)
            scale = raw_scale
            pos_mode = PositionMode.EXPLICIT

        # Transform
        transform_str = data.get("transform", "normal")
        transform_val = cls._SWAY_TRANSFORMS_INV.get(transform_str, 0)

        # Adaptive sync → VRR
        adaptive = data.get("adaptive_sync_status", "disabled")
        vrr_val = VRR.ON if adaptive == "enabled" else VRR.OFF

        return cls(
            name=data.get("name", ""),
            description=description,
            make=make,
            model=model,
            serial=serial,
            width=width,
            height=height,
            refresh_rate=refresh_rate,
            resolution_mode=ResolutionMode.EXPLICIT,
            available_modes=modes,
            x=raw_x,
            y=raw_y,
            position_mode=pos_mode,
            scale=scale,
            scale_mode=ScaleMode.EXPLICIT,
            transform=Transform(transform_val),
            enabled=enabled,
            vrr=vrr_val,
        )


# ── WorkspaceRule ────────────────────────────────────────────────────────

@dataclass
class WorkspaceRule:
    workspace: str = ""         # workspace number or name
    monitor: str = ""           # monitor name/description
    default: bool = False
    persistent: bool = False
    rounding: int = -1          # -1 = unset
    decorate: int = -1          # -1 = unset
    gapsin: int = -1
    gapsout: int = -1
    border: int = -1            # -1 = unset
    bordersize: int = -1
    on_created_empty: str = ""

    def to_hyprland_line(self, name_to_id: dict[str, str] | None = None) -> str:
        """Generate a workspace rule line for hyprland.conf."""
        parts: list[str] = [self.workspace]

        if self.monitor:
            monitor_id = (
                name_to_id[self.monitor]
                if name_to_id and self.monitor in name_to_id
                else self.monitor
            )
            parts.append(f"monitor:{monitor_id}")
        if self.default:
            parts.append("default:true")
        if self.persistent:
            parts.append("persistent:true")
        if self.rounding >= 0:
            parts.append(f"rounding:{self.rounding}")
        if self.decorate >= 0:
            parts.append(f"decorate:{self.decorate}")
        if self.gapsin >= 0:
            parts.append(f"gapsin:{self.gapsin}")
        if self.gapsout >= 0:
            parts.append(f"gapsout:{self.gapsout}")
        if self.border >= 0:
            parts.append(f"border:{self.border}")
        if self.bordersize >= 0:
            parts.append(f"bordersize:{self.bordersize}")
        if self.on_created_empty:
            parts.append(f"on-created-empty:{self.on_created_empty}")

        return "workspace=" + ", ".join(parts)

    def to_sway_line(self, name_to_id: dict[str, str] | None = None) -> str:
        """Generate a workspace assignment line for sway config.

        Sway only supports ``workspace N output NAME``.
        """
        if self.workspace and self.monitor:
            monitor_id = (
                name_to_id[self.monitor]
                if name_to_id and self.monitor in name_to_id
                else self.monitor
            )
            return f"workspace {self.workspace} output {monitor_id}"
        return ""

    def to_dict(self) -> dict:
        return asdict(self)

    @classmethod
    def from_dict(cls, d: dict) -> WorkspaceRule:
        return cls(**{k: v for k, v in d.items() if k in cls.__dataclass_fields__})

    @classmethod
    def from_hyprland_line(cls, line: str) -> WorkspaceRule | None:
        """Parse a ``workspace=...`` line from a Hyprland config file."""
        line = line.strip()
        if not line.startswith("workspace="):
            return None

        content = line[len("workspace="):]
        parts = [p.strip() for p in content.split(",")]
        if not parts:
            return None

        rule = cls(workspace=parts[0])

        _INT_FIELDS = {
            "rounding": "rounding",
            "decorate": "decorate",
            "gapsin": "gapsin",
            "gapsout": "gapsout",
            "border": "border",
            "bordersize": "bordersize",
        }

        for part in parts[1:]:
            if part.startswith("monitor:"):
                rule.monitor = part[8:]
            elif part == "default:true":
                rule.default = True
            elif part == "persistent:true":
                rule.persistent = True
            elif part.startswith("on-created-empty:"):
                rule.on_created_empty = part[17:]
            else:
                for prefix, attr in _INT_FIELDS.items():
                    if part.startswith(f"{prefix}:"):
                        try:
                            setattr(rule, attr, int(part[len(prefix) + 1:]))
                        except ValueError:
                            pass
                        break

        return rule


# ── Profile ──────────────────────────────────────────────────────────────

@dataclass
class Profile:
    name: str = ""
    monitors: list[MonitorConfig] = field(default_factory=list)
    workspace_rules: list[WorkspaceRule] = field(default_factory=list)

    @property
    def fingerprint(self) -> list[str]:
        """Sorted list of monitor descriptions for matching."""
        return sorted(m.description for m in self.monitors if m.description)

    def to_dict(self) -> dict:
        return {
            "name": self.name,
            "monitors": [m.to_dict() for m in self.monitors],
            "workspace_rules": [w.to_dict() for w in self.workspace_rules],
        }

    @classmethod
    def from_dict(cls, d: dict) -> Profile:
        return cls(
            name=d.get("name", ""),
            monitors=[MonitorConfig.from_dict(m) for m in d.get("monitors", [])],
            workspace_rules=[WorkspaceRule.from_dict(w) for w in d.get("workspace_rules", [])],
        )

    def generate_config(self, use_description: bool = False) -> str:
        """Generate the full monitors.conf content for Hyprland."""
        # Build name→identifier mapping for workspace rules and mirror references
        name_to_id: dict[str, str] = {}
        for m in self.monitors:
            if use_description and m.description:
                name_to_id[m.name] = f"desc:{m.description}"
            else:
                name_to_id[m.name] = m.name

        lines: list[str] = []
        lines.append("# Generated by Monique — https://github.com/ToRvaLDz/monique")
        lines.append("")
        for m in self.monitors:
            lines.append(m.to_hyprland_line(
                use_description=use_description, name_to_id=name_to_id,
            ))
        if self.workspace_rules:
            lines.append("")
            for w in self.workspace_rules:
                lines.append(w.to_hyprland_line(name_to_id=name_to_id))
        lines.append("")
        return "\n".join(lines)

    def generate_sway_config(self, use_description: bool = False) -> str:
        """Generate the full monitors.conf content for Sway."""
        # Build name→identifier mapping for workspace rules
        name_to_id: dict[str, str] = {}
        for m in self.monitors:
            if use_description and m.description:
                name_to_id[m.name] = f'"{m.description}"'
            else:
                name_to_id[m.name] = m.name

        blocks: list[str] = ["# Generated by Monique — https://github.com/ToRvaLDz/monique"]
        for m in self.monitors:
            blocks.append(m.to_sway_block(use_description=use_description))
        ws_lines = [
            w.to_sway_line(name_to_id=name_to_id)
            for w in self.workspace_rules
            if w.to_sway_line(name_to_id=name_to_id)
        ]
        if ws_lines:
            blocks.append("\n".join(ws_lines))
        return "\n\n".join(blocks) + "\n"

    def generate_xsetup_script(self) -> str:
        """Generate an Xsetup shell script with xrandr commands for SDDM."""
        lines = ["#!/bin/sh", "# Generated by Monique — https://github.com/ToRvaLDz/monique"]
        for m in self.monitors:
            lines.append(m.to_xrandr_cmd())
        lines.append("")
        return "\n".join(lines)


# ── Clamshell Mode ────────────────────────────────────────────────────


def apply_clamshell(monitors: list[MonitorConfig]) -> bool:
    """Disable internal displays when external monitors are present and enabled.

    Safety: does nothing if there are no enabled external monitors.
    Returns True if any monitor was changed.
    """
    internals = [m for m in monitors if m.is_internal and m.enabled]
    externals = [m for m in monitors if not m.is_internal and m.enabled]
    if not internals or not externals:
        return False
    for m in internals:
        m.enabled = False
    return True


def undo_clamshell(monitors: list[MonitorConfig]) -> bool:
    """Re-enable internal displays that were disabled.

    Returns True if any monitor was changed.
    """
    changed = False
    for m in monitors:
        if m.is_internal and not m.enabled:
            m.enabled = True
            changed = True
    return changed



"""Hyprland IPC communication via Unix sockets."""

from __future__ import annotations

import asyncio
import json
import socket
from pathlib import Path
from typing import AsyncIterator

from .models import MonitorConfig, Profile, WorkspaceRule
from .utils import (
    hyprland_runtime_dir,
    hyprland_config_dir,
    is_sway_installed,
    sway_config_dir,
    is_sddm_running,
    is_greetd_running,
    write_xsetup,
    write_greetd_monitors,
    write_text,
    backup_file,
)


class HyprlandIPC:
    """Communicate with Hyprland via its Unix socket IPC."""

    def __init__(self) -> None:
        self._runtime = hyprland_runtime_dir()

    @property
    def command_socket(self) -> Path:
        return self._runtime / ".socket.sock"

    @property
    def event_socket(self) -> Path:
        return self._runtime / ".socket2.sock"

    def _send(self, payload: bytes) -> bytes:
        """Send a raw command to the Hyprland command socket and return the response."""
        sock = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
        try:
            sock.connect(str(self.command_socket))
            sock.sendall(payload)
            chunks: list[bytes] = []
            while True:
                chunk = sock.recv(8192)
                if not chunk:
                    break
                chunks.append(chunk)
            return b"".join(chunks)
        finally:
            sock.close()

    def command(self, cmd: str) -> str:
        """Send a command and return the text response."""
        return self._send(cmd.encode()).decode(errors="replace")

    def command_json(self, cmd: str) -> list | dict:
        """Send a -j command and return parsed JSON."""
        raw = self._send(f"j/{cmd}".encode()).decode(errors="replace")
        return json.loads(raw)

    def keyword(self, key: str, value: str) -> str:
        """Send a keyword command (runtime config change)."""
        return self.command(f"keyword {key} {value}")

    def batch(self, commands: list[str]) -> str:
        """Send multiple commands as a batch."""
        joined = ";".join(commands)
        return self.command(f"[[BATCH]]{joined}")

    def reload(self) -> str:
        """Reload Hyprland configuration."""
        return self.command("reload")

    def get_monitors(self) -> list[MonitorConfig]:
        """Query all connected monitors (including disabled) as MonitorConfig list."""
        data = self.command_json("monitors all")
        return [MonitorConfig.from_hyprctl(m) for m in data]

    def get_workspaces(self) -> list[dict]:
        """Query active workspaces."""
        return self.command_json("workspaces")

    def move_workspace_to_monitor(self, workspace: str, monitor: str) -> str:
        """Move a workspace to a different monitor."""
        return self.command(f"dispatch moveworkspacetomonitor {workspace} {monitor}")

    def get_workspace_rules(self, monitors: list[MonitorConfig] | None = None) -> list[WorkspaceRule]:
        """Query workspace rules and return as WorkspaceRule list.

        Resolves ``desc:...`` monitor references to port names using *monitors*.
        """
        data = self.command_json("workspacerules")
        # Build desc→name mapping
        desc_to_name: dict[str, str] = {}
        if monitors:
            for m in monitors:
                if m.description:
                    desc_to_name[m.description] = m.name

        rules: list[WorkspaceRule] = []
        for entry in data:
            ws = entry.get("workspaceString", "")
            # Skip special workspaces
            if ws.startswith("special:"):
                continue

            monitor_raw = entry.get("monitor", "")
            if monitor_raw.startswith("desc:"):
                desc = monitor_raw[5:]
                monitor = desc_to_name.get(desc, monitor_raw)
            else:
                monitor = monitor_raw

            # gapsOut can be a list [top, right, bottom, left] or absent
            gapsout_raw = entry.get("gapsOut")
            if isinstance(gapsout_raw, list):
                gapsout = gapsout_raw[0] if gapsout_raw else -1
            elif isinstance(gapsout_raw, (int, float)):
                gapsout = int(gapsout_raw)
            else:
                gapsout = -1

            gapsin_raw = entry.get("gapsIn")
            if isinstance(gapsin_raw, list):
                gapsin = gapsin_raw[0] if gapsin_raw else -1
            elif isinstance(gapsin_raw, (int, float)):
                gapsin = int(gapsin_raw)
            else:
                gapsin = -1

            rule = WorkspaceRule(
                workspace=ws,
                monitor=monitor,
                default=entry.get("default", False),
                persistent=entry.get("persistent", False),
                rounding=entry.get("rounding", -1),
                decorate=entry.get("decorate", -1),
                gapsin=gapsin,
                gapsout=gapsout,
                border=entry.get("border", -1),
                bordersize=entry.get("borderSize", -1),
                on_created_empty=entry.get("onCreatedEmpty", ""),
            )
            rules.append(rule)
        return rules

    def apply_profile(
        self, profile: Profile, *, update_sddm: bool = True,
        update_greetd: bool = True, use_description: bool = False,
    ) -> None:
        """Write monitor config and reload Hyprland."""
        conf_dir = hyprland_config_dir()
        monitors_conf = conf_dir / "monitors.conf"

        # Backup existing
        backup_file(monitors_conf)

        # Write new config
        write_text(monitors_conf, profile.generate_config(use_description=use_description))

        # Also write Sway config if Sway is installed
        if is_sway_installed():
            sway_conf = sway_config_dir() / "monitors.conf"
            backup_file(sway_conf)
            write_text(sway_conf, profile.generate_sway_config(use_description=use_description))

        # Write SDDM Xsetup script if enabled and SDDM is present
        if update_sddm and is_sddm_running():
            write_xsetup(profile.generate_xsetup_script())

        # Write greetd sway monitors config if enabled and greetd is present
        if update_greetd and is_greetd_running():
            write_greetd_monitors(profile.generate_sway_config(use_description=use_description))

        # Reload
        self.reload()

    def apply_profile_keyword(
        self, profile: Profile, *, use_description: bool = False,
    ) -> None:
        """Apply profile via keyword commands (live, no file write)."""
        # Build name→identifier mapping
        name_to_id: dict[str, str] = {}
        for m in profile.monitors:
            if use_description and m.description:
                name_to_id[m.name] = f"desc:{m.description}"
            else:
                name_to_id[m.name] = m.name

        cmds: list[str] = []
        for m in profile.monitors:
            line = m.to_hyprland_line(
                use_description=use_description, name_to_id=name_to_id,
            )
            # strip "monitor=" prefix for keyword command
            value = line.removeprefix("monitor=")
            cmds.append(f"keyword monitor {value}")
        if cmds:
            self.batch(cmds)

    _MONITOR_EVENTS = (
        "monitoradded>>", "monitorremoved>>",
        "monitoraddedv2>>", "monitorremovedv2>>",
    )

    async def connect_event_socket(self) -> AsyncIterator[str]:
        """Connect to the event socket and yield only monitor hotplug events.

        Filters the raw event stream to yield only monitoradded/monitorremoved
        events, so callers don't need to filter themselves.
        """
        reader, _ = await asyncio.open_unix_connection(str(self.event_socket))
        while True:
            line = await reader.readline()
            if not line:
                break
            event = line.decode(errors="replace").strip()
            if any(event.startswith(e) for e in self._MONITOR_EVENTS):
                yield event
